<?php

namespace Drupal\workbench_approver_access\EventSubscriber;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\workbench_access\Entity\AccessSchemeInterface;
use Drupal\workbench_access\RoleSectionStorageInterface;
use Drupal\workbench_access\UserSectionStorageInterface;
use Drupal\workbench_access\WorkbenchAccessManagerInterface;
use Drupal\workbench_approver\ApproverEvents;
use Drupal\workbench_approver\Event\GetPotentialApproversEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Overlays Workbench Access rules on approval assignments.
 */
class WorkbenchApproverAccessSubscriber implements EventSubscriberInterface {

  /**
   * The Workbench Access manager service.
   *
   * @var \Drupal\workbench_access\WorkbenchAccessManager
   */
  protected $manager;

  /**
   * The user section storage service.
   *
   * @var \Drupal\workbench_access\UserSectionStorageInterface
   */
  protected $userSectionStorage;

  /**
   * Role section storage service.
   *
   * @var \Drupal\workbench_access\RoleSectionStorageInterface
   */
  protected $roleSectionStorage;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Module config.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Constructs a new WorkbenchApproverAccessSubscriber.
   *
   * @param \Drupal\workbench_access\WorkbenchAccessManagerInterface $manager
   *   The Workbench Access hierarchy manager.
   * @param \Drupal\workbench_access\UserSectionStorageInterface $user_section_storage
   *   The user section storage service.
   * @param \Drupal\workbench_access\RoleSectionStorageInterface $role_section_storage
   *   Role section storage.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Config factory.
   */
  public function __construct(
    WorkbenchAccessManagerInterface $manager,
    UserSectionStorageInterface $user_section_storage,
    RoleSectionStorageInterface $role_section_storage,
    EntityTypeManagerInterface $entity_type_manager,
    ConfigFactoryInterface $config_factory,
  ) {
    $this->manager = $manager;
    $this->userSectionStorage = $user_section_storage;
    $this->roleSectionStorage = $role_section_storage;
    $this->entityTypeManager = $entity_type_manager;
    $this->configFactory = $config_factory;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events = [];
    $events[ApproverEvents::WORKBENCH_APPROVER_GET_POTENTIAL_APPROVERS][] = ['modifyResults'];
    return $events;
  }

  /**
   * Sets the potential approvers for this content.
   *
   * This event subscriber overlays Workbench Access rules on top
   * of the Workbench Approver module, removing the ability to
   * assign an approver who is not a member of the editorial access
   * group that the entity is in.
   *
   * If the entity is not assigned to an access group, we use the
   * 'deny_on_empty' setting for Workbench Access.
   *
   * @param \Drupal\workbench_approver\Event\GetPotentialApproversEvent $event
   *   The event, which has arguments and an approver list.
   *
   * @see \Drupal\workbench_approver\Services\ApproverManager::getPotentialApproverIds()
   * @see \Drupal\workbench_approver\Event\GetPotentialApproversEvent::setPotentialApprovers()
   */
  public function modifyResults(GetPotentialApproversEvent $event): void {
    $arguments = $event->getArguments();
    // @phpstan-ignore-next-line
    $entity = $this->entityTypeManager->getStorage($arguments['content_entity_type_id'])
      ->loadRevision($arguments['content_entity_revision_id']);

    // If no entity loads, we cannot do anything.
    if ($entity === NULL) {
      return;
    }

    $allowed_editors = NULL;
    $config = $this->configFactory->get('workbench_access.settings');
    $access_schemes = $this->entityTypeManager->getStorage('access_scheme')->loadMultiple();

    foreach ($access_schemes as $scheme) {
      assert($scheme instanceof AccessSchemeInterface);
      $access_scheme = $scheme->getAccessScheme();
      // Find the scheme that applies to this entity.
      if ($access_scheme->applies($entity->getEntityTypeId(), $entity->bundle())) {
        $entity_sections = $access_scheme->getEntityValues($entity);
        // If the entity has no records, check the setting for that condition.
        if (count($entity_sections) === 0) {
          // Ignore workbench_access in this case.
          if ($config->get('deny_on_empty') === FALSE) {
            return;
          }
          // Deny all in this case.
          else {
            $allowed_editors = [];
          }
        }
        else {
          $existing_editors = [];
          $editors = [];
          $count = count($entity_sections) - 1;
          foreach ($entity_sections as $index => $section_id) {
            // We have to get the tree and all parents of the section.
            $ids = $this->getParents($scheme, $section_id);
            $editors[$section_id] = [];
            foreach ($ids as $id) {
              $editors[$section_id] += array_keys($this->userSectionStorage->getEditors($scheme, $id));
              $editors[$section_id] += $this->getUsersByRole($this->roleSectionStorage->getRoles($scheme, $id));
            }

            // If there are multiple sections, the users must be in all.
            if ($count === 0) {
              $existing_editors = current($editors);
            }
            elseif ($index === $count) {
              // First, compile a full list.
              foreach ($editors as $editor) {
                $existing_editors = array_unique(array_merge($existing_editors, $editor));
              }
              // Now ensure each item is in all lists.
              foreach ($existing_editors as $key => $uid) {
                foreach ($editors as $editor_list) {
                  // @phpstan-ignore-next-line
                  if (!in_array($uid, $editor_list)) {
                    unset($existing_editors[$key]);
                  }
                }
              }
            }
          }
          // Account for users who can 'bypass workbench_access'.
          $existing_editors = array_merge($existing_editors, $this->getBypassUsers());
          $allowed_editors = array_unique($existing_editors);
        }
      }
    }

    // If we have a return value, we must enforce it.
    if ($allowed_editors !== NULL) {
      $approvers = $event->getPotentialApproverIds();
      $event->setPotentialApprovers(array_intersect($allowed_editors, $approvers));
    }
  }

  /**
   * Returns the parent items of an editorial section.
   *
   * @param \Drupal\workbench_access\Entity\AccessSchemeInterface $scheme
   *   The active access scheme.
   * @param string|int $id
   *   The section id.
   *
   * @return array
   *   An array of section ids, including the child and its parents.
   */
  public function getParents(AccessSchemeInterface $scheme, $id) {
    $ids = [$id];
    $tree = $scheme->getAccessScheme()->getTree();
    foreach ($tree as $info) {
      if (isset($info[$id]['parents'])) {
        $ids = array_merge($ids, $info[$id]['parents']);
      }
    }
    return $ids;
  }

  /**
   * Finds all users within the roles assigned to the entity.
   *
   * @param array $rids
   *   An array of role ids.
   *
   * @return array
   *   An array of user uids.
   *
   * @see \Drupal\workbench_access\UserSectionStorage::getPotentialEditorsRoles()
   */
  public function getUsersByRole(array $rids): array {
    $storage = $this->entityTypeManager->getStorage('user');

    $query = $storage->getAggregateQuery()->accessCheck(FALSE);
    $query->condition('status', 1);
    if (count($rids) === 0) {
      return [];
    }
    // If authenticated role included, we return all.
    elseif (!in_array(AccountInterface::AUTHENTICATED_ROLE, $rids, TRUE)) {
      $query->condition('roles', $rids, 'IN');
    }

    $results = $query->groupBy('uid')->execute();

    return array_column($results, 'uid');
  }

  /**
   * Gets all users with the 'bypass workbench access' permission.
   *
   * @return array
   *   An array of user ids.
   */
  public function getBypassUsers() {
    $storage = $this->entityTypeManager->getStorage('user_role');
    $roles = $storage->loadMultiple();
    $rids = [];
    foreach ($roles as $role) {
      if ($role->hasPermission('bypass workbench access')) {
        $rids[] = $role->id();
      }
    }

    return $this->getUsersByRole($rids);
  }

}
