<?php

namespace Drupal\Tests\workbench_approver_access\Functional;

use Drupal\workbench_access\WorkbenchAccessManagerInterface;

/**
 * Tests for access to approval request creation / viewing.
 *
 * @group workbench_approver
 */
class ApproveFormAccessTest extends ApproverAccessTestBase {

  /**
   * Tests for access to approval forms.
   *
   * Note that `add approver` is embedded in the
   * node tab and not otherwise accessible. Tests for that
   * are covered in AssignAccessTest.
   *
   * Other form paths are "guessable" if the user knows
   * the approval request id, and should have access checks.
   */
  public function testApprovalFormAccess() {
    $values = [
      'type' => 'page',
      'title' => 'Page node',
      'moderation_state' => 'published',
    ];
    $node1 = $this->drupalCreateNode($values);
    $revision1 = $this->createRevision($node1);

    // Node is not in workbench_access.
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node1);
    $this->assertEmpty($entity_sections);

    // Testing paths.
    $base_path = '/node/' . $node1->id();
    $this->drupalGet($base_path);

    // Create an approval request.
    $approval = $this->addApprovalRequest($this->editorUser, $revision1);

    // We need to ensure that the forms are not accessible.
    // Those forms live at /approver/{approver}/edit etc.
    // They are normally accessed via AJAX, but need to be checked
    // here as well. See workbench_approver_access_approver_access().
    // Basic rules:
    // - Only the assigner/owner can edit or delete.
    // - The assigner/owner and the approver can approve/reject.
    // - Any role with the view permission can see notes.
    // When workbench_approver_access is enabled:
    // - Enforce the above for content using Workbench Access so
    // that users must meet those conditions and be in the proper
    // editorial section.
    // Test case 1a:
    // What if a node is not assigned to Workbench Access
    // And 'deny_on_empty' is FALSE (allow).
    $users = [
      NULL,
      $this->editorUser,
      $this->adminUser,
      $this->anotherEditorUser,
      $this->observerUser,
    ];

    $paths = [
      '/approver/' . $approval->id() . '/edit' => [
        $this->adminUser->id(),
      ],
      '/approver/' . $approval->id() . '/approve' => [
        $this->editorUser->id(),
        $this->adminUser->id(),
      ],
      '/approver/' . $approval->id() . '/reject' => [
        $this->editorUser->id(),
        $this->adminUser->id(),
      ],
      '/approver/' . $approval->id() . '/delete' => [
        $this->adminUser->id(),
      ],
      '/approver/' . $approval->id() . '/notes'  => [
        $this->editorUser->id(),
        $this->anotherEditorUser->id(),
        $this->adminUser->id(),
        $this->observerUser->id(),
      ],
    ];
    $this->assertApprovalPathAccess($paths, $users);

    // Test case 1b:
    // What if a node is not assigned to Workbench Access
    // And 'deny_on_empty' is TRUE (deny)
    $this->assertDenyOnEmpty();

    $paths = [
      '/approver/' . $approval->id() . '/edit' => [],
      '/approver/' . $approval->id() . '/approve' => [],
      '/approver/' . $approval->id() . '/reject' => [],
      '/approver/' . $approval->id() . '/delete' => [],
      '/approver/' . $approval->id() . '/notes'  => [],
    ];
    $this->assertApprovalPathAccess($paths, $users);

    // Test case 2:
    // What if a node and a user are assigned to a section (allow)
    $this->assertSectionAssigned();

    // Create a node that is assigned to a section.
    $node2 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      'moderation_state' => 'published',
      WorkbenchAccessManagerInterface::FIELD_NAME => $this->baseTerm->id(),
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node2);
    $this->assertNotEmpty($entity_sections);
    $revision2 = $this->createRevision($node2);

    $approval2 = $this->addApprovalRequest($this->editorUser, $revision2);

    // This is a bit of an odd case, since we did not grant
    // access to the `adminUser` who created the request.
    // Still, this tests the logic accurately. See test 3b
    // for the updated access.
    $paths2 = [
      '/approver/' . $approval2->id() . '/edit' => [],
      '/approver/' . $approval2->id() . '/approve' => [
        $this->editorUser->id(),
      ],
      '/approver/' . $approval2->id() . '/reject' => [
        $this->editorUser->id(),
      ],
      '/approver/' . $approval2->id() . '/delete' => [],
      '/approver/' . $approval2->id() . '/notes'  => [
        $this->editorUser->id(),
      ],
    ];
    $this->assertApprovalPathAccess($paths2, $users);

    // Test case 3:
    // What if a node and a role are assigned to a section (allow)
    // We still use approval2 here, since access assignments changed.
    $this->assertAddSectionToEditorRole();

    // Recheck the other user's request.
    $paths3 = [
      '/approver/' . $approval2->id() . '/edit' => [],
      '/approver/' . $approval2->id() . '/approve' => [
        $this->editorUser->id(),
      ],
      '/approver/' . $approval2->id() . '/reject' => [
        $this->editorUser->id(),
      ],
      '/approver/' . $approval2->id() . '/delete' => [],
      '/approver/' . $approval2->id() . '/notes'  => [
        $this->editorUser->id(),
        $this->anotherEditorUser->id(),
      ],
    ];
    $this->assertApprovalPathAccess($paths3, $users);

    $approval3 = $this->addApprovalRequest($this->anotherEditorUser, $revision2);

    // Check this user's request.
    $paths3 = [
      '/approver/' . $approval3->id() . '/edit' => [],
      '/approver/' . $approval3->id() . '/approve' => [
        $this->anotherEditorUser->id(),
      ],
      '/approver/' . $approval3->id() . '/reject' => [
        $this->anotherEditorUser->id(),
      ],
      '/approver/' . $approval3->id() . '/delete' => [],
      '/approver/' . $approval3->id() . '/notes'  => [
        $this->editorUser->id(),
        $this->anotherEditorUser->id(),
      ],
    ];
    $this->assertApprovalPathAccess($paths3, $users);

    // Test case 3b:
    // Add the admin user and check again.
    $user_storage = $this->container->get('workbench_access.user_section_storage');
    $user_storage->addUser($this->scheme, $this->adminUser, [$this->baseTerm->id()]);

    // Check after adding the admin.
    $paths3 = [
      '/approver/' . $approval3->id() . '/edit' => [
        $this->adminUser->id(),
      ],
      '/approver/' . $approval3->id() . '/approve' => [
        $this->adminUser->id(),
        $this->anotherEditorUser->id(),
      ],
      '/approver/' . $approval3->id() . '/reject' => [
        $this->adminUser->id(),
        $this->anotherEditorUser->id(),
      ],
      '/approver/' . $approval3->id() . '/delete' => [
        $this->adminUser->id(),
      ],
      '/approver/' . $approval3->id() . '/notes'  => [
        $this->adminUser->id(),
        $this->editorUser->id(),
        $this->anotherEditorUser->id(),
      ],
    ];
    $this->assertApprovalPathAccess($paths3, $users);

    // Test case 4:
    // What if a node is assigned to a subsection of the user's (allow)
    $node4 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      'moderation_state' => 'published',
      WorkbenchAccessManagerInterface::FIELD_NAME => $this->subTerm->id(),
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node4);
    $this->assertNotEmpty($entity_sections);
    $revision4 = $this->createRevision($node4);
    $approval4 = $this->addApprovalRequest($this->anotherEditorUser, $revision4);

    // Should be the same as test 3b.
    $paths4 = [
      '/approver/' . $approval4->id() . '/edit' => [
        $this->adminUser->id(),
      ],
      '/approver/' . $approval4->id() . '/approve' => [
        $this->adminUser->id(),
        $this->anotherEditorUser->id(),
      ],
      '/approver/' . $approval4->id() . '/reject' => [
        $this->adminUser->id(),
        $this->anotherEditorUser->id(),
      ],
      '/approver/' . $approval4->id() . '/delete' => [
        $this->adminUser->id(),
      ],
      '/approver/' . $approval4->id() . '/notes'  => [
        $this->adminUser->id(),
        $this->editorUser->id(),
        $this->anotherEditorUser->id(),
      ],
    ];
    $this->assertApprovalPathAccess($paths4, $users);

    // Test case 5:
    // What if a node is in a section with no users (deny)
    $node5 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      'moderation_state' => 'published',
      WorkbenchAccessManagerInterface::FIELD_NAME => $this->unrelatedTerm->id(),
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node5);
    $this->assertNotEmpty($entity_sections);
    $revision5 = $this->createRevision($node5);

    // Create an approval request (which can't really happen).
    $approval5 = $this->addApprovalRequest($this->editorUser, $revision5);
    $paths5 = [
      '/approver/' . $approval5->id() . '/edit' => [],
      '/approver/' . $approval5->id() . '/approve' => [],
      '/approver/' . $approval5->id() . '/reject' => [],
      '/approver/' . $approval5->id() . '/delete' => [],
      '/approver/' . $approval5->id() . '/notes'  => [],
    ];
    $this->assertApprovalPathAccess($paths5, $users);

    // Test case 6a:
    // What if a node is in multiple sections
    // And a user is only in one section (deny)
    $node6 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      'moderation_state' => 'published',
      WorkbenchAccessManagerInterface::FIELD_NAME => [$this->unrelatedTerm->id(), $this->baseTerm->id()],
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node6);
    $this->assertNotEmpty($entity_sections);
    $revision6 = $this->createRevision($node6);
    $approval6 = $this->addApprovalRequest($this->editorUser, $revision6);

    $paths6 = [
      '/approver/' . $approval6->id() . '/edit' => [],
      '/approver/' . $approval6->id() . '/approve' => [],
      '/approver/' . $approval6->id() . '/reject' => [],
      '/approver/' . $approval6->id() . '/delete' => [],
      '/approver/' . $approval6->id() . '/notes'  => [],
    ];
    $this->assertApprovalPathAccess($paths6, $users);

    // Test case 6b:
    // What if a node is in multiple sections
    // And a user is in all sections (allow)
    $user_storage->addUser($this->scheme, $this->editorUser, [$this->unrelatedTerm->id()]);
    $expected = [$this->editorUser->id()];
    $existing_users = $user_storage->getEditors($this->scheme, $this->unrelatedTerm->id());
    $this->assertEquals($expected, array_keys($existing_users));

    $paths6 = [
      '/approver/' . $approval6->id() . '/edit' => [],
      '/approver/' . $approval6->id() . '/approve' => [
        $this->editorUser->id(),
      ],
      '/approver/' . $approval6->id() . '/reject' => [
        $this->editorUser->id(),
      ],
      '/approver/' . $approval6->id() . '/delete' => [],
      '/approver/' . $approval6->id() . '/notes'  => [
        $this->editorUser->id(),
      ],
    ];
    $this->assertApprovalPathAccess($paths6, $users);

    // Test case 6c:
    // What if a user can 'bypass workbench access' (allow)
    // For this test, we add in the 'observer' role for clarity.
    $permissions = [
      'use workbench access',
      'use workbench approver',
      'bypass workbench access',
    ];
    $role_storage = $this->container->get('entity_type.manager')->getStorage('user_role');
    $role = $role_storage->load('observer');
    foreach ($permissions as $permission) {
      $role->grantPermission($permission);
    }
    $role->save();

    $paths6 = [
      '/approver/' . $approval6->id() . '/edit' => [],
      '/approver/' . $approval6->id() . '/approve' => [
        $this->editorUser->id(),
      ],
      '/approver/' . $approval6->id() . '/reject' => [
        $this->editorUser->id(),
      ],
      '/approver/' . $approval6->id() . '/delete' => [],
      '/approver/' . $approval6->id() . '/notes'  => [
        $this->editorUser->id(),
        $this->observerUser->id(),
      ],
    ];
    $this->assertApprovalPathAccess($paths6, $users);

  }

}
