<?php

namespace Drupal\Tests\workbench_approver_access\Functional;

use Drupal\workbench_access\WorkbenchAccessManagerInterface;

/**
 * Tests for lists of available approvers.
 *
 * @group workbench_approver
 */
class ApproverAccessSubscriberTest extends ApproverAccessTestBase {

  /**
   * Tests that our user list is as expected.
   *
   * This is a very extensive test to cover lots of cases.
   *
   * 1. What if a node is not assigned to Workbench Access
   *   a. And 'deny_on_empty' is FALSE (allow)
   *   b. And 'deny_on_empty' is TRUE  (deny)
   * 2. What if a node and a user are assigned to a section (allow)
   * 3. What if a node and a role are assigned to a section (allow)
   * 4. What if a node is assigned to a subsection of the user's (allow)
   * 5. What if a node is in a section with no users (deny)
   * 6. What if a node is in multiple sections
   *   a. And a user is only in one section (deny)
   *   b. And a user is in all sections (allow)
   * 7. What if a user can 'bypass workbench access' (allow)
   *
   * @covers \Drupal\workbench_approver_access\EventSubscriber\WorkbenchApproverAccessSubscriber
   */
  public function testGetPotentialApprovers() {
    // Test case 1a:
    // What if a node is not assigned to Workbench Access
    // And 'deny_on_empty' is FALSE (allow)
    $expected_uids = [
      $this->adminUser->id(),
      $this->editorUser->id(),
      $this->anotherEditorUser->id(),
      $this->workflowAdminUser->id(),
    ];

    $not_expected_uids = [
      $this->observerUser->id(),
    ];

    $values = [
      'type' => 'page',
      'title' => 'Page node',
      'moderation_state' => 'published',
    ];
    $node1 = $this->drupalCreateNode($values);
    $this->createRevision($node1);

    // Node is not in workbench_access.
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node1);
    $this->assertEmpty($entity_sections);

    $arguments = [
      'content_entity_id' => $node1->id(),
      'content_entity_revision_id' => $node1->getRevisionId(),
      'content_entity_type_id' => 'node',
      'workflow_id' => $this->workflow->id(),
    ];

    $this->checkApprovers($arguments, $expected_uids, $not_expected_uids);

    // Test case 1b:
    // What if a node is not assigned to Workbench Access
    // And 'deny_on_empty' is TRUE (deny)
    $this->assertDenyOnEmpty();
    $potential_approvers = $this->approverManager->getPotentialApproverIds($arguments);
    $this->assertEquals([], $potential_approvers, 'No users are expected');

    // Test case 2:
    // What if a node and a user are assigned to a section (allow)
    $this->assertSectionAssigned();

    // Create a node that is assigned to a section.
    $node2 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      WorkbenchAccessManagerInterface::FIELD_NAME => $this->baseTerm->id(),
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node2);
    $this->assertNotEmpty($entity_sections);
    $this->createRevision($node2);

    $arguments = [
      'content_entity_id' => $node2->id(),
      'content_entity_revision_id' => $node2->getRevisionId(),
      'content_entity_type_id' => 'node',
      'workflow_id' => $this->workflow->id(),
    ];

    $expected_uids = [
      $this->editorUser->id(),
    ];

    $not_expected_uids = [
      $this->adminUser->id(),
      $this->observerUser->id(),
      $this->workflowAdminUser->id(),
    ];

    $this->checkApprovers($arguments, $expected_uids, $not_expected_uids);

    // Test case 3:
    // What if a node and a role are assigned to a section (allow)
    $this->assertAddSectionToEditorRole();

    $expected_uids = [
      $this->editorUser->id(),
      $this->anotherEditorUser->id(),
    ];

    $not_expected_uids = [
      $this->adminUser->id(),
      $this->observerUser->id(),
      $this->workflowAdminUser->id(),
    ];

    $this->checkApprovers($arguments, $expected_uids, $not_expected_uids);

    // Test case 4:
    // What if a node is assigned to a subsection of the user's (allow)
    $node3 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      WorkbenchAccessManagerInterface::FIELD_NAME => $this->subTerm->id(),
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node3);
    $this->assertNotEmpty($entity_sections);
    $this->createRevision($node3);

    $arguments = [
      'content_entity_id' => $node3->id(),
      'content_entity_revision_id' => $node3->getRevisionId(),
      'content_entity_type_id' => 'node',
      'workflow_id' => $this->workflow->id(),
    ];

    $expected_uids = [
      $this->editorUser->id(),
      $this->anotherEditorUser->id(),
    ];

    $not_expected_uids = [
      $this->adminUser->id(),
      $this->observerUser->id(),
      $this->workflowAdminUser->id(),
    ];

    $this->checkApprovers($arguments, $expected_uids, $not_expected_uids);

    // Test case 5:
    // What if a node is in a section with no users (deny)
    $node4 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      WorkbenchAccessManagerInterface::FIELD_NAME => $this->unrelatedTerm->id(),
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node4);
    $this->assertNotEmpty($entity_sections);
    $this->createRevision($node4);

    $expected_uids = [];

    $not_expected_uids = [
      $this->adminUser->id(),
      $this->editorUser->id(),
      $this->observerUser->id(),
      $this->workflowAdminUser->id(),
      $this->anotherEditorUser->id(),
    ];

    $arguments = [
      'content_entity_id' => $node4->id(),
      'content_entity_revision_id' => $node4->getRevisionId(),
      'content_entity_type_id' => 'node',
      'workflow_id' => $this->workflow->id(),
    ];

    $this->checkApprovers($arguments, $expected_uids, $not_expected_uids);

    // Test case 6a:
    // What if a node is in multiple sections
    // And a user is only in one section (deny)
    $node5 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      WorkbenchAccessManagerInterface::FIELD_NAME => [$this->unrelatedTerm->id(), $this->baseTerm->id()],
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node5);
    $this->assertNotEmpty($entity_sections);
    $this->createRevision($node5);

    $expected_uids = [];

    $not_expected_uids = [
      $this->adminUser->id(),
      $this->editorUser->id(),
      $this->observerUser->id(),
      $this->workflowAdminUser->id(),
      $this->anotherEditorUser->id(),
    ];

    $arguments = [
      'content_entity_id' => $node5->id(),
      'content_entity_revision_id' => $node5->getRevisionId(),
      'content_entity_type_id' => 'node',
      'workflow_id' => $this->workflow->id(),
    ];

    $this->checkApprovers($arguments, $expected_uids, $not_expected_uids);

    // Test case 6b:
    // What if a node is in multiple sections
    // And a user is in all sections (allow)
    $user_storage = $this->container->get('workbench_access.user_section_storage');
    $user_storage->addUser($this->scheme, $this->editorUser, [$this->unrelatedTerm->id()]);
    $expected = [$this->editorUser->id()];
    $existing_users = $user_storage->getEditors($this->scheme, $this->unrelatedTerm->id());
    $this->assertEquals($expected, array_keys($existing_users));

    $expected_uids = [
      $this->editorUser->id(),
    ];

    $not_expected_uids = [
      $this->adminUser->id(),
      $this->observerUser->id(),
      $this->workflowAdminUser->id(),
      $this->anotherEditorUser->id(),
    ];

    $this->checkApprovers($arguments, $expected_uids, $not_expected_uids);

    // Test case 7:
    // What if a user can 'bypass workbench access' (allow)
    // For this test, we add in the 'observer' role for clarity.
    $permissions = [
      'use workbench access',
      'use workbench approver',
      'bypass workbench access',
    ];
    $role_storage = $this->container->get('entity_type.manager')->getStorage('user_role');
    $role = $role_storage->load('observer');
    foreach ($permissions as $permission) {
      $role->grantPermission($permission);
    }
    $role->save();

    $expected_uids = [
      $this->editorUser->id(),
      $this->observerUser->id(),
    ];

    $not_expected_uids = [
      $this->adminUser->id(),
      $this->workflowAdminUser->id(),
      $this->anotherEditorUser->id(),
    ];

    $this->checkApprovers($arguments, $expected_uids, $not_expected_uids);

    // Remove the permission and test again.
    $role->revokePermission('bypass workbench access')->save();
    $expected_uids = [
      $this->editorUser->id(),
    ];

    $not_expected_uids = [
      $this->adminUser->id(),
      $this->observerUser->id(),
      $this->workflowAdminUser->id(),
      $this->anotherEditorUser->id(),
    ];

    $this->checkApprovers($arguments, $expected_uids, $not_expected_uids);
  }

}
