<?php

namespace Drupal\Tests\workbench_approver_access\Functional;

use Drupal\workbench_access\WorkbenchAccessManagerInterface;

/**
 * Tests for access to approval request creation / viewing.
 *
 * @todo This may be better as a Kernel test.
 *
 * @group workbench_approver
 */
class AssignAccessTest extends ApproverAccessTestBase {

  /**
   * Tests for access to the Approve tab.
   */
  public function testApproveTabAccess() {
    $values = [
      'type' => 'page',
      'title' => 'Page node',
      'moderation_state' => 'published',
    ];
    $node1 = $this->drupalCreateNode($values);
    $this->createRevision($node1);

    // Node is not in workbench_access.
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node1);
    $this->assertEmpty($entity_sections);

    // Testing paths.
    $base_path = '/node/' . $node1->id();
    $tab_path = $base_path . '/approver';
    $latest_path = $base_path . '/latest';

    // Test case 1a:
    // What if a node is not assigned to Workbench Access
    // And 'deny_on_empty' is FALSE (allow)
    $users = [
      NULL,
      $this->editorUser,
      $this->anotherEditorUser,
      $this->adminUser,
      $this->observerUser,
    ];

    foreach ($users as $user) {
      $status_code = 403;
      if ($user !== NULL) {
        $this->drupalLogin($user);
        $status_code = 200;
      }
      // There is a published, accessible version.
      $this->drupalGet($base_path);
      $this->assertSession()->statusCodeEquals(200);
      // Approve tab for all users.
      $this->drupalGet($tab_path);
      $this->assertSession()->statusCodeEquals($status_code);
      // The latest tab should be visible to admin users.
      $status_code = ($user !== NULL && $user->id() === $this->adminUser->id()) ? 200 : 403;
      $this->drupalGet($latest_path);
      $this->assertSession()->statusCodeEquals($status_code);
    }

    // Test case 1b:
    // What if a node is not assigned to Workbench Access
    // And 'deny_on_empty' is TRUE (deny)
    $this->assertDenyOnEmpty();

    foreach ($users as $user) {
      $status_code = 403;
      if ($this->loggedInUser) {
        $this->drupalLogout();
      }
      if ($user !== NULL) {
        $this->drupalLogin($user);
      }
      // There is a published, accessible version.
      $this->drupalGet($base_path);
      $this->assertSession()->statusCodeEquals(200);
      // Approve tab for all users.
      $this->drupalGet($tab_path);
      $this->assertSession()->statusCodeEquals($status_code);
      // The latest tab should not be visible to anyone.
      $this->drupalGet($latest_path);
      $this->assertSession()->statusCodeEquals($status_code);
    }

    // Test case 2:
    // What if a node and a user are assigned to a section (allow)
    $this->assertSectionAssigned();

    // Create a node that is assigned to a section.
    $node2 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      'moderation_state' => 'published',
      WorkbenchAccessManagerInterface::FIELD_NAME => $this->baseTerm->id(),
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node2);
    $this->assertNotEmpty($entity_sections);
    $revision2 = $this->createRevision($node2);

    $this->addApprovalRequest($this->editorUser, $revision2);

    // Testing paths.
    $base_path2 = '/node/' . $node2->id();
    $tab_path2 = $base_path2 . '/approver';
    $latest_path2 = $base_path2 . '/latest';

    foreach ($users as $user) {
      $status_code = 403;
      if ($this->loggedInUser) {
        $this->drupalLogout();
      }
      if ($user !== NULL) {
        $this->drupalLogin($user);
        if ($user->id() === $this->editorUser->id()) {
          $status_code = 200;
        }
      }
      // There is a published, accessible version.
      $this->drupalGet($base_path2);
      $this->assertSession()->statusCodeEquals(200);
      // Approve tab for all users.
      $this->drupalGet($tab_path2);
      $this->assertSession()->statusCodeEquals($status_code);
      // The latest tab should be visible to the editor user.
      $this->drupalGet($latest_path2);
      $this->assertSession()->statusCodeEquals($status_code);
    }

    // Test case 3:
    // What if a node and a role are assigned to a section (allow)
    $this->assertAddSectionToEditorRole();

    foreach ($users as $user) {
      $status_code = 403;
      if ($this->loggedInUser) {
        $this->drupalLogout();
      }
      if ($user !== NULL) {
        $this->drupalLogin($user);
        if ($user->id() === $this->editorUser->id() || $user->id() === $this->anotherEditorUser->id()) {
          $status_code = 200;
        }
      }
      // There is a published, accessible version.
      $this->drupalGet($base_path2);
      $this->assertSession()->statusCodeEquals(200);
      // Approve tab for all users.
      $this->drupalGet($tab_path2);
      $this->assertSession()->statusCodeEquals($status_code);
      // The latest tab should be visible to the editor user.
      // The anotherEditorUser cannot see it because they need
      // to be an approver, too.
      $this->drupalGet($latest_path2);
      $status_code = ($user !== NULL && $user->id() === $this->editorUser->id()) ? 200 : 403;
      $this->assertSession()->statusCodeEquals($status_code);
    }

    // Test case 4:
    // What if a node is assigned to a subsection of the user's (allow)
    $node3 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      'moderation_state' => 'published',
      WorkbenchAccessManagerInterface::FIELD_NAME => $this->subTerm->id(),
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node3);
    $this->assertNotEmpty($entity_sections);
    $revision3 = $this->createRevision($node3);
    $this->addApprovalRequest($this->editorUser, $revision3);

    // Testing paths.
    $base_path3 = '/node/' . $node3->id();
    $tab_path3 = $base_path3 . '/approver';
    $latest_path3 = $base_path3 . '/latest';

    foreach ($users as $user) {
      $status_code = 403;
      if ($this->loggedInUser) {
        $this->drupalLogout();
      }
      if ($user !== NULL) {
        $this->drupalLogin($user);
        if ($user->id() === $this->editorUser->id() || $user->id() === $this->anotherEditorUser->id()) {
          $status_code = 200;
        }
      }
      // There is a published, accessible version.
      $this->drupalGet($base_path3);
      $this->assertSession()->statusCodeEquals(200);
      // Approve tab for all users.
      $this->drupalGet($tab_path3);
      $this->assertSession()->statusCodeEquals($status_code);
      // The latest tab should be visible to the editor user.
      // The anotherEditorUser cannot see it because they need
      // to be an approver, too.
      $this->drupalGet($latest_path3);
      $status_code = ($user !== NULL && $user->id() === $this->editorUser->id()) ? 200 : 403;
      $this->assertSession()->statusCodeEquals($status_code);
    }

    // Test case 5:
    // What if a node is in a section with no users (deny)
    $node4 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      'moderation_state' => 'published',
      WorkbenchAccessManagerInterface::FIELD_NAME => $this->unrelatedTerm->id(),
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node4);
    $this->assertNotEmpty($entity_sections);
    $this->createRevision($node4);

    // Testing paths.
    $base_path4 = '/node/' . $node4->id();
    $tab_path4 = $base_path4 . '/approver';
    $latest_path4 = $base_path4 . '/latest';

    foreach ($users as $user) {
      if ($this->loggedInUser) {
        $this->drupalLogout();
      }
      $status_code = 403;
      if ($user !== NULL) {
        $this->drupalLogin($user);
      }
      // There is a published, accessible version.
      $this->drupalGet($base_path4);
      $this->assertSession()->statusCodeEquals(200);
      // Approve tab for all users.
      $this->drupalGet($tab_path4);
      $this->assertSession()->statusCodeEquals($status_code);
      // The latest tab should not be visible.
      $this->drupalGet($latest_path4);
      $this->assertSession()->statusCodeEquals($status_code);
    }

    // Test case 6a:
    // What if a node is in multiple sections
    // And a user is only in one section (deny)
    $node5 = $this->createNode([
      'type' => 'page',
      'title' => 'bar',
      'moderation_state' => 'published',
      WorkbenchAccessManagerInterface::FIELD_NAME => [$this->unrelatedTerm->id(), $this->baseTerm->id()],
    ]);
    $entity_sections = $this->scheme->getAccessScheme()->getEntityValues($node5);
    $this->assertNotEmpty($entity_sections);
    $revision5 = $this->createRevision($node5);
    $this->addApprovalRequest($this->editorUser, $revision5);

    // Testing paths.
    $base_path5 = '/node/' . $node5->id();
    $tab_path5 = $base_path5 . '/approver';
    $latest_path5 = $base_path5 . '/latest';

    foreach ($users as $user) {
      $status_code = 403;
      if ($this->loggedInUser) {
        $this->drupalLogout();
      }
      if ($user !== NULL) {
        $this->drupalLogin($user);
      }
      // There is a published, accessible version.
      $this->drupalGet($base_path5);
      $this->assertSession()->statusCodeEquals(200);
      // Approve tab for all users.
      $this->drupalGet($tab_path5);
      $this->assertSession()->statusCodeEquals($status_code);
      // The latest tab should not be visible.
      $this->drupalGet($latest_path5);
      $this->assertSession()->statusCodeEquals($status_code);
    }

    // Test case 6b:
    // What if a node is in multiple sections
    // And a user is in all sections (allow)
    $user_storage = $this->container->get('workbench_access.user_section_storage');
    $user_storage->addUser($this->scheme, $this->editorUser, [$this->unrelatedTerm->id()]);
    $expected = [$this->editorUser->id()];
    $existing_users = $user_storage->getEditors($this->scheme, $this->unrelatedTerm->id());
    $this->assertEquals($expected, array_keys($existing_users));

    foreach ($users as $user) {
      $status_code = 403;
      if ($this->loggedInUser) {
        $this->drupalLogout();
      }
      if ($user !== NULL) {
        $this->drupalLogin($user);
        if ($user->id() === $this->editorUser->id()) {
          $status_code = 200;
        }
      }
      // There is a published, accessible version.
      $this->drupalGet($base_path5);
      $this->assertSession()->statusCodeEquals(200);
      // Approve tab for all users.
      $this->drupalGet($tab_path5);
      $this->assertSession()->statusCodeEquals($status_code);
      // The latest tab should be visible to the editor user.
      $this->drupalGet($latest_path5);
      $this->assertSession()->statusCodeEquals($status_code);
    }

    // Test case 6c:
    // What if a user can 'bypass workbench access' (allow)
    // For this test, we add in the 'observer' role for clarity.
    $permissions = [
      'use workbench access',
      'use workbench approver',
      'bypass workbench access',
    ];
    $this->addPermissionsToRole($permissions, 'observer');

    foreach ($users as $user) {
      $status_code = 403;
      if ($this->loggedInUser) {
        $this->drupalLogout();
      }
      if ($user !== NULL) {
        $this->drupalLogin($user);
        if ($user->id() === $this->editorUser->id() || $user->id() === $this->observerUser->id()) {
          $status_code = 200;
        }
      }
      // There is a published, accessible version.
      $this->drupalGet($base_path5);
      $this->assertSession()->statusCodeEquals(200);
      // Approve tab for all users.
      $this->drupalGet($tab_path5);
      $this->assertSession()->statusCodeEquals($status_code);
      // The latest tab should only be visible to the assigned user.
      $this->drupalGet($latest_path5);
      $status_code = ($user !== NULL && $user->id() === $this->editorUser->id()) ? 200 : 403;
      $this->assertSession()->statusCodeEquals($status_code);
    }

  }

}
