<?php

namespace Drupal\Tests\workbench_approver_lock\Functional;

use Drupal\Tests\workbench_approver\Functional\WorkbenchApproverTestBase;
use Drupal\user\UserInterface;

/**
 * Tests for access to approval pages and tabs.
 *
 * @group workbench_approver
 */
class WorkbenchApproverLockEntityAccessTest extends WorkbenchApproverTestBase {

  /**
   * Editor user.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $anotherEditorUser;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'workbench_approver',
    'workbench_approver_lock',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    // We need the Editor role to access Revisions in this test since the test
    // is about locking the access to Edit and Revisions for not approvers.
    $this->addPermissionsToRole(['view all revisions', 'view any unpublished content'], 'editor');
  }

  /**
   * Test access to the 'Edit' and 'Revisions' tab.
   *
   * @covers ::workbench_approver_lock_node_access
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException|\Drupal\Core\Entity\EntityStorageException
   */
  public function testEditAndRevisionsTabAccess() {
    $approved_node_non_required = $this->drupalCreateNode([
      'type' => 'approved',
      'title' => 'Approved node',
      'moderation_state' => 'draft',
      'uid' => $this->adminUser->id(),
    ]);

    $approved_node_required = $this->drupalCreateNode([
      'type' => 'approved',
      'title' => 'Approved node',
      'moderation_state' => 'draft',
      'uid' => $this->adminUser->id(),
    ]);

    // Testing paths.
    $approved_non_required_base_path = '/node/' . $approved_node_non_required->id();
    $approved_non_required_edit_path = $approved_non_required_base_path . '/edit';
    $approved_non_required_revisions_path = $approved_non_required_base_path . '/revisions';
    $approved_required_base_path = '/node/' . $approved_node_required->id();
    $approved_required_edit_path = $approved_required_base_path . '/edit';
    $approved_required_revisions_path = $approved_required_base_path . '/revisions';

    // Testing before adding Approval Requests.
    // Since there are no approval requests, all access is allowed.
    $this->drupalLogin($this->editorUser);

    $this->assertStatusCode(200, [
      $approved_non_required_base_path,
      $approved_non_required_edit_path,
      $approved_non_required_revisions_path,
    ]);
    $this->assertStatusCode(200, [
      $approved_required_base_path,
      $approved_required_edit_path,
      $approved_required_revisions_path,
    ]);

    // Add non-blocking and blocking requests to the nodes with another editor
    // as approver to test for access on both cases.
    $this->addApprovalRequest($this->anotherEditorUser, $approved_node_non_required, ['required' => FALSE]);
    $this->addApprovalRequest($this->anotherEditorUser, $approved_node_required, ['required' => TRUE]);

    // User 'editorUser' is not the approver so should not be able to edit or
    // see the revisions of the node.
    $this->assertStatusCode(200, [
      $approved_non_required_base_path,
      $approved_non_required_edit_path,
      $approved_non_required_revisions_path,
    ]);
    $this->assertStatusCode(200, [$approved_required_base_path]);
    $this->assertStatusCode(403, [
      $approved_required_edit_path,
      $approved_required_revisions_path,
    ]);
  }

  /**
   * Tests the given paths for a given http status.
   *
   * This function helps to reduce the number of repeated lines when several
   * paths are expected to get the same Status Code (usually 200 or 403).
   *
   * @param int $status
   *   The status code that the paths are expected to have.
   * @param array $paths
   *   The paths to be asserted.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   */
  private function assertStatusCode(int $status, array $paths): void {
    foreach ($paths as $path) {
      $this->drupalGet($path);
      $this->assertSession()->statusCodeEquals($status);
    }
  }

}
