<?php

namespace Drupal\workbench_approver\Access;

use Drupal\content_moderation\Access\LatestRevisionCheck;
use Drupal\content_moderation\ModerationInformationInterface;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\RevisionableInterface;
use Drupal\Core\Routing\Access\AccessInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\workbench_approver\Entity\ApproverInterface;
use Symfony\Component\Routing\Route;

/**
 * Adds assigned reviewers access to the latest revision tab.
 */
class ApproveLatestRevisionCheck extends LatestRevisionCheck {

  /**
   * The decorated access check.
   *
   * @var \Drupal\Core\Routing\Access\AccessInterface
   */
  protected $accessCheck;

  /**
   * The moderation information service.
   *
   * @var \Drupal\content_moderation\ModerationInformationInterface
   */
  protected $moderationInfo;

  /**
   * ApproveLatestRevisionCheck constructor.
   *
   * @param \Drupal\Core\Routing\Access\AccessInterface $access_check
   *   Latest revision access check to decorate.
   * @param \Drupal\content_moderation\ModerationInformationInterface $moderation_information
   *   The moderation information service.
   */
  public function __construct(AccessInterface $access_check, ModerationInformationInterface $moderation_information) {
    $this->accessCheck = $access_check;
    $this->moderationInfo = $moderation_information;
  }

  /**
   * {@inheritdoc}
   */
  public function access(Route $route, RouteMatchInterface $route_match, AccountInterface $account) {
    $entity = $this->loadEntity($route, $route_match);
    // Only run this check on moderated entities.
    if ($this->moderationInfo->hasPendingRevision($entity)) {
      /** @var \Drupal\Core\Access\AccessResultInterface $access_result */
      $access_result = $this->accessCheck->access($route, $route_match, $account);
      if (!$access_result->isAllowed() && $this->checkAssignments($entity, $account)) {
        $access_result = AccessResult::allowedIfHasPermission($account, 'use workbench approver');
      }
      return $access_result->addCacheableDependency($entity)->addCacheableDependency($account);
    }

    return AccessResult::forbidden('No pending revision for moderated entity.')->addCacheableDependency($entity)->addCacheableDependency($account);
  }

  /**
   * Checks if a user is assigned to approve the entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity being checked.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The user making the request.
   *
   * @return bool
   *   TRUE if the user is assigned as a reviewer.
   */
  public function checkAssignments(EntityInterface $entity, AccountInterface $account): bool {
    $return = FALSE;

    // Ensure that we have an entity in moderation.
    assert($entity instanceof RevisionableInterface);
    $workflow = $this->moderationInfo->getWorkflowForEntity($entity);
    if ($workflow !== NULL) {
      // @todo Can we inject this service?
      // @phpstan-ignore-next-line
      $storage = \Drupal::entityTypeManager()->getStorage('approver');

      $values = [
        'content_entity_type_id' => $entity->getEntityTypeId(),
        'content_entity_id' => $entity->id(),
        'content_entity_revision_id' => $entity->getRevisionId(),
        'approver_uid' => $account->id(),
        'workflow_id' => $workflow->id(),
        'langcode' => $entity->language()->getId(),
      ];

      $assigned = $storage->loadByProperties($values);

      // Assignments (not approvals) are active if > 0.
      if (count($assigned) > 0) {
        $record = current($assigned);
        assert($record instanceof ApproverInterface);
        if ($record->getStatus() > ApproverInterface::APPROVAL_RESCINDED) {
          $return = TRUE;
        }
      }
    }

    return $return;
  }

}
