<?php

namespace Drupal\workbench_approver;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityListBuilder;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\workbench_approver\Entity\ApproverInterface;
use Drupal\workbench_approver\Services\ApproverManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines a class to build a listing of approval record entities.
 *
 * @see \Drupal\workbench_approver\Entity\Approver
 */
class ApproverListBuilder extends EntityListBuilder {

  /**
   * The approver manager service.
   *
   * @var \Drupal\workbench_approver\Services\ApproverManagerInterface
   */
  protected $approverManager;

  /**
   * {@inheritdoc}
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    return new static(
      $entity_type,
      $container->get('entity_type.manager')->getStorage($entity_type->id()),
      $container->get('workbench_approver.manager')
    );
  }

  /**
   * Constructs a new ApproverListBuilder object.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   * @param \Drupal\Core\Entity\EntityStorageInterface $storage
   *   The entity storage class.
   * @param \Drupal\workbench_approver\Services\ApproverManagerInterface $approver_manager
   *   The approver manager.
   */
  public function __construct(EntityTypeInterface $entity_type, EntityStorageInterface $storage, ApproverManagerInterface $approver_manager) {
    parent::__construct($entity_type, $storage);
    $this->approverManager = $approver_manager;
  }

  /**
   * Gets this list's default operations.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity the operations are for.
   *
   * @return array
   *   The array structure is identical to the return value of
   *   self::getOperations().
   */
  protected function getDefaultOperations(EntityInterface $entity) {
    $operations = [];

    assert($entity instanceof ApproverInterface);
    $status = $entity->getStatus();
    // View link to latest revision of the parent entity.
    $parent_entity = $entity->getApprovalRevision();
    if ($parent_entity instanceof EntityInterface && $parent_entity->access('view')) {
      $url = $this->approverManager->getRevisionUrl($parent_entity);
      $operations['view'] = [
        'title' => $this->t('Show revision'),
        'weight' => -15,
        'url' => $this->ensureDestination($url),
      ];
    }
    // View notes if present.
    if ($entity->access('view') && $entity->hasLinkTemplate('notes-form')) {
      $notes = $entity->get('notes')->value;
      if ($notes !== NULL) {
        $operations['notes'] = [
          'title' => $this->t('View notes'),
          'weight' => -10,
          'attributes' => [
            'class' => ['use-ajax'],
            'data-dialog-type' => 'modal',
            'data-dialog-options' => Json::encode([
              'width' => 880,
            ]),
          ],
          'url' => $this->ensureDestination($entity->toUrl('notes-form')),
        ];
      }
    }
    // Approval / rejection options.
    if ($entity->access('update') && $status === ApproverInterface::APPROVAL_PENDING) {
      if ($entity->hasLinkTemplate('approve-form')) {
        $operations['approve'] = [
          'title' => $this->t('Approve revision'),
          'weight' => 0,
          'attributes' => [
            'class' => ['use-ajax'],
            'data-dialog-type' => 'modal',
            'data-dialog-options' => Json::encode([
              'width' => 880,
            ]),
          ],
          'url' => $this->ensureDestination($entity->toUrl('approve-form')),
        ];
      }
      if ($entity->hasLinkTemplate('reject-form')) {
        $operations['reject'] = [
          'title' => $this->t('Request changes'),
          'weight' => 10,
          'attributes' => [
            'class' => ['use-ajax'],
            'data-dialog-type' => 'modal',
            'data-dialog-options' => Json::encode([
              'width' => 880,
            ]),
          ],
          'url' => $this->ensureDestination($entity->toUrl('reject-form')),
        ];
      }
    }
    // We use `delete` to indicate people who can do more than approve/deny.
    if ($entity->access('delete')) {
      if ($entity->hasLinkTemplate('edit-form')) {
        $operations['change'] = [
          'title' => $this->t('Edit request'),
          'weight' => 20,
          'attributes' => [
            'class' => ['use-ajax'],
            'data-dialog-type' => 'modal',
            'data-dialog-options' => Json::encode([
              'width' => 880,
            ]),
          ],
          'url' => $this->ensureDestination($entity->toUrl('edit-form')),
        ];
      }
      if ($entity->hasLinkTemplate('delete-form')) {
        $operations['delete'] = [
          'title' => $this->t('Delete request'),
          'weight' => 50,
          'attributes' => [
            'class' => ['use-ajax'],
            'data-dialog-type' => 'modal',
            'data-dialog-options' => Json::encode([
              'width' => 880,
            ]),
          ],
          'url' => $this->ensureDestination($entity->toUrl('delete-form')),
        ];
      }
    }

    return $operations;
  }

}
