<?php

namespace Drupal\workbench_approver;

use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Entity\RevisionableInterface;
use Drupal\Core\Entity\Sql\SqlContentEntityStorage;

/**
 * Defines the storage handler class for approval requests.
 *
 * This extends the base storage class, adding required special handling for
 * approval request entities.
 *
 * Primarily, our create() method needs to check for
 * an existing entity that has been marked as deprecated
 * since we don't delete things due to audit trails.
 */
class ApproverStorage extends SqlContentEntityStorage implements ApproverStorageInterface {

  /**
   * {@inheritdoc}
   */
  public function create(array $values = []) {
    $bundle = $this->getBundleFromValues($values);
    $entity_class = $this->getEntityClass($bundle);
    // @todo Decide what to do if preCreate() tries to change the bundle.
    // @see https://www.drupal.org/project/drupal/issues/3230792
    $entity_class::preCreate($this, $values);

    $existing_entity = $this->checkExistingEntity($values);

    // If we are restoring a "deleted" request, make a new revision.
    // We do this for audit trails. Requests are never deleted.
    if ($existing_entity !== NULL) {
      $entity = clone $existing_entity;
      assert($entity instanceof RevisionableInterface);
      $entity->setNewRevision();
      foreach ($values as $key => $value) {
        // @phpstan-ignore-next-line
        $entity->set($key, $value);
      }
    }
    else {
      // Assign a new UUID if there is none yet.
      if ($this->uuidService !== NULL && $this->uuidKey !== FALSE && !isset($values[$this->uuidKey])) {
        assert($this->uuidService instanceof UuidInterface);
        $values[$this->uuidKey] = $this->uuidService->generate();
      }

      $entity = $this->doCreate($values);

      // @todo Do we need to invoke hooks here?
      if (!isset($values['id'])) {
        $entity->enforceIsNew();

        $entity->postCreate($this);

        // Modules might need to add or change the data initially held
        // by the new entity object, for instance to fill-in default values.
        $this->invokeHook('create', $entity);
      }

    }

    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function checkExistingEntity(array $values) {
    // We only lookup by certain values here.
    $entities = $this->loadByProperties([
      'content_entity_type_id' => $values['content_entity_type_id'],
      'content_entity_id' => $values['content_entity_id'],
      'content_entity_revision_id' => $values['content_entity_revision_id'],
      'approver_uid' => $values['approver_uid'],
      'workflow_id' => $values['workflow_id'],
      'langcode' => $values['langcode'],
    ]);
    if (count($entities) > 0) {
      $entity = current($entities);
    }
    return $entity ?? NULL;
  }

}
