<?php

namespace Drupal\workbench_approver\Controller;

use Drupal\content_moderation\ModerationInformationInterface;
use Drupal\content_moderation\StateTransitionValidationInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\workbench_approver\Services\ApproverManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Controller for the approvals on a node.
 *
 * @todo Expand to other entity types later.
 */
class ApproverController extends ControllerBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The content moderation information service.
   *
   * @var \Drupal\content_moderation\ModerationInformationInterface
   */
  protected $moderationInformation;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * The entity repository service.
   *
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected $entityRepository;

  /**
   * The approver manager service.
   *
   * @var \Drupal\workbench_approver\Services\ApproverManagerInterface
   */
  protected $approverManager;

  /**
   * The form builder service.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;

  /**
   * The state transition validator.
   *
   * @var \Drupal\workbench_approver\Services\ApproverStateTransitionValidation
   */
  protected $transitionValidation;

  /**
   * The entity being viewed.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $entity;

  /**
   * The current workflow for the entity.
   *
   * @var \Drupal\workflows\WorkflowInterface
   */
  protected $workflow;

  /**
   * The curren moderation state for the entity.
   *
   * @var \Drupal\workflows\StateInterface
   */
  protected $moderationState;

  /**
   * Indicates that we show summary data to this user.
   *
   * @var bool
   */
  protected $showSummary = FALSE;

  /**
   * Creates an ApproverController object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\content_moderation\ModerationInformationInterface $content_moderation_info
   *   The content moderation information service.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The current user.
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository
   *   The entity repository.
   * @param \Drupal\workbench_approver\Services\ApproverManagerInterface $approver_manager
   *   The approver manager.
   * @param \Drupal\Core\Form\FormBuilderInterface $form_builder
   *   The form builder.
   * @param \Drupal\content_moderation\StateTransitionValidationInterface $transition_validation
   *   The state transition validator.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    ModerationInformationInterface $content_moderation_info,
    AccountProxyInterface $current_user,
    EntityRepositoryInterface $entity_repository,
    ApproverManagerInterface $approver_manager,
    FormBuilderInterface $form_builder,
    StateTransitionValidationInterface $transition_validation,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->moderationInformation = $content_moderation_info;
    $this->currentUser = $current_user;
    $this->entityRepository = $entity_repository;
    $this->approverManager = $approver_manager;
    $this->formBuilder = $form_builder;
    $this->transitionValidation = $transition_validation;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('content_moderation.moderation_information'),
      $container->get('current_user'),
      $container->get('entity.repository'),
      $container->get('workbench_approver.manager'),
      $container->get('form_builder'),
      $container->get('workbench_approver.state_transition_validation')
    );
  }

  /**
   * Admin page callback for approvals by content.
   *
   * These are used because Views menu handling is suspect.
   *
   * @return array
   *   A Drupal render array.
   */
  public function adminByContent(): array {
    $content = [];

    $content['approvals'] = views_embed_view('workbench_approver_admin_by_content', 'content_list', 'node');

    return $content;
  }

  /**
   * Admin page callback for approvals by user.
   *
   * These are used because Views menu handling is suspect.
   *
   * @return array
   *   A Drupal render array.
   */
  public function adminByUser(): array {
    $content = [];

    $content['approvals'] = views_embed_view('workbench_approver_admin_by_user', 'user_list', 'node');

    return $content;
  }

  /**
   * The _title_callback for the page that shows content approvals.
   *
   * @todo The route is keyed to nodes right now. If we support
   * additional entity types, the argument will change.
   *
   * @param \Drupal\Core\Entity\EntityInterface $node
   *   The current node.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The page title.
   */
  public function title(EntityInterface $node): TranslatableMarkup {
    return $this->t('Approvals for %title', [
      '%title' => $this->entityRepository->getTranslationFromContext($node)->label(),
    ]);
  }

  /**
   * Content callback for the node 'approve' tab.
   *
   * This tab displays information based on user permissions
   * and approval request status, so we break it into parts.
   *
   * @todo The route is keyed to nodes right now. If we support
   * additional entity types, the argument will change.
   *
   * @param \Drupal\Core\Entity\EntityInterface $node
   *   The node revision under approval.
   *
   * @return array
   *   A Drupal render array.
   */
  public function approvalTab(EntityInterface $node): array {
    $content = [];
    $this->entity = $node;

    // Ensure the entity is under moderation.
    $this->checkModerationState();

    // We should never hit this condition due to route access checks.
    if ($this->moderationState === NULL) {
      $content['#markup'] = $this->t('This content is not being moderated and is not subject to approvals.');
      return $content;
    }

    $this->getAssignerMessage($content);
    $this->getReviewerMessage($content);
    $this->getSummary($content);
    $this->getActiveApprovals($content);
    $this->getAddApproverForm($content);
    $this->getPriorApprovals($content);
    $this->getRevisionHistory($content);

    return $content;
  }

  /**
   * Confirms the validated moderation state of an entity.
   *
   * This method also sets $this->workflow and $this->moderationState.
   */
  private function checkModerationState() {
    $this->moderationState = NULL;

    $this->workflow = $this->moderationInformation->getWorkflowForEntity($this->entity);
    if ($this->workflow !== NULL) {
      $moderation_state_id = $this->entity->get('moderation_state')->getString();
      if ($this->workflow->getTypePlugin()->hasState($moderation_state_id)) {
        $this->moderationState = $this->workflow->getTypePlugin()->getState($moderation_state_id);
      }
    }
  }

  /**
   * Prepares the help text for users who can assign approvals.
   *
   * @param array $content
   *   The content array, passed by reference.
   */
  private function getAssignerMessage(array &$content) {
    // Messaging to users.
    if ($this->approverManager->isAssigner()) {
      $content['assigner'] = [
        '#markup' => $this->t('Each revision requires approval before the
          content can move to the next state. Revisions without assigned reviewers can be
          promoted by any user with the proper permission.'),
        '#prefix' => '<p>',
        '#suffix' => '<p>',
      ];
      // This person is an assigner.
      $content['reviewer_assigner'] = [
        '#markup' => $this->t('You may assign approvals for this content.'),
        '#prefix' => '<p>',
        '#suffix' => '<p>',
      ];

      $this->showSummary = TRUE;
    }
  }

  /**
   * Prepares the help text for users who can see approvals.
   *
   * @param array $content
   *   The content array, passed by reference.
   */
  private function getReviewerMessage(array &$content) {
    $active = $this->approverManager->getApprovalRequestByUser($this->entity, $this->currentUser, TRUE);
    if ($active !== NULL) {
      $content['reviewer'] = [
        '#markup' => $this->t('Your review has been requested on the <a href="@link">current revision</a>.', [
          '@link' => $this->approverManager->getRevisionUrl($this->entity)->toString(),
        ]),
        '#prefix' => '<p>',
        '#suffix' => '<p>',
      ];
      $this->showSummary = TRUE;
    }
    else {
      $prior = $this->approverManager->getApprovalRequestByUser($this->entity, $this->currentUser, FALSE);
      if ($prior !== NULL) {
        $content['reviewer'] = [
          '#markup' => $this->t('There are no pending approvals for you on this content.'),
          '#prefix' => '<p>',
          '#suffix' => '<p>',
        ];

        $this->showSummary = TRUE;
      }
      // This person is just a viewer.
      elseif (!$this->approverManager->isAssigner()) {
        $content['reviewer'] = [
          '#markup' => $this->t('You may view the approvals for this content.'),
          '#prefix' => '<p>',
          '#suffix' => '<p>',
        ];
      }
    }
  }

  /**
   * Prepares the summary text for users who can assign approvals.
   *
   * We do not want to leak revision labels to users who cannot
   * view revisions. $this->showSummary should handle that.
   *
   * @param array $content
   *   The content array, passed by reference.
   */
  private function getSummary(array &$content) {
    // Summary of the current state -- only shown to people allowed to see it.
    // @todo The showSummary logic is not encapsulated properly in one call.
    if ($this->showSummary) {
      $content['summary'] = [
        '#theme' => 'item_list',
        '#items' => [
          $this->t('Current revision state: %state', ['%state' => $this->moderationState->label()]),
          $this->t('Current revision id: %revision', ['%revision' => $this->entity->getRevisionId()]),
          $this->t('Current revision: <a href="@link">@title</a>', [
            '@link' => $this->approverManager->getRevisionUrl($this->entity)->toString(),
            '@title' => $this->entity->label(),
          ]),
        ],
      ];
    }
  }

  /**
   * Prepares the table of active approvals.
   *
   * Note that access here is handled by the view and route.
   *
   * @param array $content
   *   The content array, passed by reference.
   */
  private function getActiveApprovals(array &$content) {
    $content['active_header']['#markup'] = '<h3>' . $this->t('Active approvals') . '</h3>';
    // Show null text or the view.
    if ($this->entity->isPublished()) {
      $content['active_approvals']['#markup'] = $this->t('The current version is published and cannot be approved further.');
    }
    // If the latest revision cannot transition, approvers cannot be assigned.
    elseif (!$this->transitionValidation->isApprovable($this->entity)) {
      $content['active_approvals']['#markup'] = $this->t('No current approvals are required for this content.');
      // Help text for admins.
      if ($this->currentUser->hasPermission('administer workflows')) {
        $content['active_approvals'] = [
          '#markup' => $this->t('Approvals may only be assigned when new revisions will not change the current state of the content.'),
          '#prefix' => '<p>',
          '#suffix' => '<p>',
        ];
        $content['active_approvals_help'] = [
          '#markup' => $this->t('The current %state moderation state does not support this action and may need <a href="@link">a new content transition</a>.',
            [
              '%state' => $this->moderationState->label(),
              '@link' => $this->workflow->toUrl('edit-form')->toString(),
            ]),
          '#prefix' => '<p>',
          '#suffix' => '<p>',
        ];
      }
    }
    else {
      $content['active_approvals'] = views_embed_view(
        'workbench_approver_default_tables',
        'active_approvals',
        $this->entity->id(),
        $this->entity->getRevisionId(),
        $this->entity->getEntityTypeId()
      );
    }
  }

  /**
   * Prepares the form to assign new approvers.
   *
   * @param array $content
   *   The content array, passed by reference.
   */
  private function getAddApproverForm(array &$content) {
    // Admins can add new reviewers.
    if ($this->approverManager->isAssigner() && !$this->entity->isPublished() &&
        $this->transitionValidation->isApprovable($this->entity)) {
      // Pass the nid, vid, workflow_id, state id, and langcode.
      // @phpstan-ignore-next-line
      $content['add_approver_form'] = $this->formBuilder->getForm(
        'Drupal\workbench_approver\Form\AddApproverForm',
        $this->entity->id(),
        $this->entity->getRevisionId(),
        $this->workflow->id(),
        $this->moderationState->id(),
        $this->currentUser->id(),
        $this->entity->language()->getId()
      );
    }
  }

  /**
   * Prepares the view for prior approvals.
   *
   * Note that access here is handled by the view and route.
   *
   * @param array $content
   *   The content array, passed by reference.
   */
  private function getPriorApprovals(array &$content) {
    // Show the view for prior revision approvals.
    $content['prior_header']['#markup'] = '<h3>' . $this->t('Prior approvals') . '</h3>';
    $content['prior_approvals'] = views_embed_view(
      'workbench_approver_default_tables',
      'prior_approvals',
      $this->entity->id(),
      $this->entity->getRevisionId(),
      $this->entity->getEntityTypeId()
    );
  }

  /**
   * Prepares the revision history for users with access.
   *
   * @param array $content
   *   The content array, passed by reference.
   */
  private function getRevisionHistory(array &$content) {
    // Show our revision history view if allowed.
    if ($this->currentUser->hasPermission('view all revisions')) {
      $content['revision_header']['#markup'] = '<h3>' . $this->t('Revision history') . '</h3>';
      $content['revision_history'] = views_embed_view('workbench_approver_revision_history', 'history', $this->entity->id());
    }
  }

}
