<?php

namespace Drupal\workbench_approver\Entity;

use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\RevisionableStorageInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\user\EntityOwnerTrait;

/**
 * Defines the Workbench Approver entity.
 *
 * @ContentEntityType(
 *   id = "approver",
 *   label = @Translation("Content approval record"),
 *   label_singular = @Translation("content approval record"),
 *   label_plural = @Translation("content approval records"),
 *   label_count = @PluralTranslation(
 *     singular = "@count content approval record",
 *     plural = "@count content approval records"
 *   ),
 *   handlers = {
 *     "storage" = "Drupal\workbench_approver\ApproverStorage",
 *     "access" = "Drupal\workbench_approver\ApproverAccessControlHandler",
 *     "views_data" = "Drupal\workbench_approver\ApproverViewsData",
 *     "list_builder" = "Drupal\workbench_approver\ApproverListBuilder",
 *     "form" = {
 *       "default" = "Drupal\workbench_approver\Form\UpdateApproverForm",
 *       "delete" = "Drupal\workbench_approver\Form\DeleteApproverForm",
 *       "edit" = "Drupal\workbench_approver\Form\UpdateApproverForm",
 *       "approve" = "Drupal\workbench_approver\Form\ApproveForm",
 *       "reject" = "Drupal\workbench_approver\Form\RejectForm",
 *       "notes" = "Drupal\workbench_approver\Form\NotesForm",
 *     },
 *   },
 *   base_table = "workbench_approver",
 *   revision_table = "workbench_approver_revision",
 *   data_table = "workbench_approver_field_data",
 *   revision_data_table = "workbench_approver_field_revision",
 *   translatable = TRUE,
 *   static_cache = TRUE,
 *   entity_keys = {
 *     "id" = "id",
 *     "revision" = "revision_id",
 *     "uuid" = "uuid",
 *     "uid" = "uid",
 *     "owner" = "uid",
 *     "approver_uid" = "approver_uid",
 *     "langcode" = "langcode",
 *   },
 *   links = {
 *     "canonical" = "/approver/{approver}",
 *     "delete-form" = "/approver/{approver}/delete",
 *     "edit-form" = "/approver/{approver}/edit",
 *     "approve-form" = "/approver/{approver}/approve",
 *     "reject-form" = "/approver/{approver}/reject",
 *     "notes-form" = "/approver/{approver}/notes",
 *     "create" = "/approver",
 *   }
 * )
 */
class Approver extends ContentEntityBase implements ApproverInterface {

  use EntityOwnerTrait;

  /**
   * {@inheritdoc}
   */
  public static function statusList(): array {
    return [
      ApproverInterface::APPROVAL_RESCINDED => t('Rescinded'),
      ApproverInterface::APPROVAL_INVALIDATED => t('Invalidated'),
      ApproverInterface::APPROVAL_OUTDATED => t('Outdated'),
      ApproverInterface::APPROVAL_PENDING => t('Pending'),
      ApproverInterface::APPROVAL_GRANTED => t('Approved'),
      ApproverInterface::APPROVAL_REJECTED => t('Requested changes'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function statusName(int $status): TranslatableMarkup {
    $statuses = static::statusList();

    return $statuses[$status];
  }

  /**
   * {@inheritdoc}
   */
  public static function statusValue(string $status): int {
    $statuses = static::statusList();
    $statuses = array_flip($statuses);

    return $statuses[$status];
  }

  /**
   * {@inheritdoc}
   */
  public function isBlocking(): bool {
    return ($this->isRequired() && ($this->getStatus() === ApproverInterface::APPROVAL_PENDING || $this->getStatus() === ApproverInterface::APPROVAL_REJECTED));
  }

  /**
   * {@inheritdoc}
   */
  public function isPending(): bool {
    return ($this->getStatus() === ApproverInterface::APPROVAL_PENDING);
  }

  /**
   * {@inheritdoc}
   */
  public function isRejected(): bool {
    return ($this->getStatus() === ApproverInterface::APPROVAL_REJECTED);
  }

  /**
   * {@inheritdoc}
   */
  public function isRequired(): bool {
    return $this->get('required')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function isActive(): bool {
    return $this->getStatus() >= ApproverInterface::APPROVAL_PENDING;
  }

  /**
   * {@inheritdoc}
   */
  public function getStatus(): int {
    return $this->get('status')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function getStatusName(): TranslatableMarkup {
    return static::statusName($this->get('status')->value);
  }

  /**
   * {@inheritdoc}
   */
  public function getApproverId(): int {
    $value = current($this->get('approver_uid')->getValue());

    return intval($value['target_id']);
  }

  /**
   * {@inheritdoc}
   */
  public function getApproverLabel(): ?string {
    $value = current($this->get('approver_uid')->getValue());
    // phpcs:ignore
    $storage = \Drupal::entityTypeManager()->getStorage('user');
    $user = $storage->load($value['target_id']);

    return $user?->label();
  }

  /**
   * {@inheritdoc}
   */
  public function getOwnerLabel(): ?string {
    $value = intval($this->get('uid')->getString());
    // phpcs:ignore
    $storage = \Drupal::entityTypeManager()->getStorage('user');
    $user = $storage->load($value);

    return $user?->label();
  }

  /**
   * {@inheritdoc}
   */
  public function getAssignerId(): int {
    $value = current($this->get('uid')->getValue());

    return intval($value['target_id']);
  }

  /**
   * {@inheritdoc}
   */
  public function getContentEntityId(): int {
    $value = $this->get('content_entity_id')->getString();
    return intval($value);
  }

  /**
   * {@inheritdoc}
   */
  public function getContentEntityRevisionId(): int {
    $value = $this->get('content_entity_revision_id')->getString();
    return intval($value);
  }

  /**
   * {@inheritdoc}
   */
  public function getAssignerLabel() {
    return $this->getOwner()->label();
  }

  /**
   * {@inheritdoc}
   */
  public function getApprovalRevision() {
    $entity = NULL;
    $entity_type_id = $this->get('content_entity_type_id')->value;
    $revision_id = $this->get('content_entity_revision_id')->value;
    // phpcs:ignore
    $storage = \Drupal::entityTypeManager()->getStorage($entity_type_id);

    // @todo Error handling.
    if ($storage instanceof RevisionableStorageInterface) {
      $entity = $storage->loadRevision($revision_id);
    }

    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type);
    $fields += static::ownerBaseFieldDefinitions($entity_type);

    $fields['workflow_id'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Workflow ID'))
      ->setDescription(t('The workflow the moderation state is in.'))
      ->setSetting('target_type', 'workflow')
      ->setRequired(TRUE)
      ->setRevisionable(TRUE);

    $fields['moderation_state'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Moderation state'))
      ->setDescription(t('The moderation state of the referenced content.'))
      ->setRequired(TRUE)
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE);

    $fields['content_entity_type_id'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Content entity type ID'))
      ->setDescription(t('The ID of the content entity type this moderation state is for.'))
      ->setRequired(TRUE)
      ->setSetting('max_length', EntityTypeInterface::ID_MAX_LENGTH)
      ->setRevisionable(TRUE);

    $fields['content_entity_id'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Content entity ID'))
      ->setDescription(t('The ID of the content entity this moderation state is for.'))
      ->setRequired(TRUE)
      ->setRevisionable(TRUE);

    $fields['content_entity_revision_id'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Content entity revision ID'))
      ->setDescription(t('The revision ID of the content entity this moderation state is for.'))
      ->setRequired(TRUE)
      ->setRevisionable(TRUE);

    assert($fields['uid'] instanceof BaseFieldDefinition);
    $fields['uid']
      ->setLabel(t('User'))
      ->setDescription(t('The user id of the person who made the assignment (the entity owner).'))
      ->setRevisionable(TRUE);

    $fields['approver_uid'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Approver'))
      ->setDescription(t('The username of an assigned approver.'))
      ->setRevisionable(TRUE)
      ->setSetting('target_type', 'user')
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'user_name',
        'weight' => 0,
      ])
      ->setDisplayOptions('form', [
        'type' => 'entity_reference_autocomplete',
        'weight' => 5,
        'settings' => [
          'match_operator' => 'CONTAINS',
          'size' => '60',
          'placeholder' => '',
        ],
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['status'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Status'))
      ->setDescription(t('The status of the approval request.'))
      ->setRevisionable(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'select',
        'label' => 'hidden',
        'weight' => 120,
      ])
      ->setDisplayConfigurable('form', FALSE);

    $fields['required'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Required'))
      ->setDescription(t('Indicates the approval is required.'))
      ->setRevisionable(TRUE)
      ->setTranslatable(TRUE)
      ->setDefaultValue(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'boolean_checkbox',
        'settings' => [
          'display_label' => TRUE,
        ],
        'weight' => 15,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['notes'] = BaseFieldDefinition::create('string_long')
      ->setLabel(t('Notes'))
      ->setRequired(FALSE)
      ->setDescription(t('Notes appended to the approval request.'))
      ->setRevisionable(TRUE)
      ->setTranslatable(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'string_textarea',
        'weight' => 0,
        'settings' => [
          'rows' => 8,
        ],
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'type' => 'string',
        'weight' => 0,
        'label' => 'above',
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['deadline'] = BaseFieldDefinition::create('timestamp')
      ->setLabel(t('Due date'))
      ->setDescription(t('The date that the assignment is due.'))
      ->setRevisionable(TRUE)
      ->setTranslatable(TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'date',
        'weight' => 0,
      ])
      ->setDisplayOptions('form', [
        'type' => 'date',
        'weight' => 10,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Requested'))
      ->setDescription(t('The date that the assignment was created.'))
      ->setRevisionable(TRUE)
      ->setTranslatable(TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'timestamp',
        'weight' => 0,
      ])
      ->setDisplayOptions('form', [
        'type' => 'datetime_timestamp',
        'weight' => 10,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Updated'))
      ->setDescription(t('The date that the assignment was last edited.'))
      ->setRevisionable(TRUE)
      ->setTranslatable(TRUE);

    return $fields;
  }

}
