<?php

namespace Drupal\workbench_approver\Form;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\workbench_approver\Entity\ApproverInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form to add a reviewer to a node revision.
 *
 * @todo Currently, this form is not an entity form due
 * to the dependencies on other values.
 */
class AddApproverForm extends FormBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  public $entityTypeManager;

  /**
   * The datetime service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  public $dateTime;

  /**
   * Creates an AddApproverForm object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $datetime
   *   The datetime service.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    TimeInterface $datetime,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->dateTime = $datetime;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('datetime.time')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'approver';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // Get information passed from the controller.
    [$nid, $vid, $workflow_id, $moderation_state, $uid, $langcode] = $form_state->getBuildInfo()['args'];

    $form['step'] = [
      '#type' => 'value',
      '#value' => $form_state->getValue('step') ?? 0,
    ];

    $form['actions'] = ['#type' => 'actions'];

    $request_time = $this->dateTime->getRequestTime();

    // These elements cannot be edited directly.
    $form['workflow_id'] = ['#type' => 'value', '#value' => $workflow_id];
    $form['moderation_state'] = ['#type' => 'value', '#value' => $moderation_state];
    $form['content_entity_type_id'] = ['#type' => 'value', '#value' => 'node'];
    $form['content_entity_id'] = ['#type' => 'value', '#value' => $nid];
    $form['content_entity_revision_id'] = ['#type' => 'value', '#value' => $vid];
    $form['langcode'] = ['#type' => 'value', '#value' => $langcode];

    // @todo These might go in the create() method.
    $form['uid'] = ['#type' => 'value', '#value' => $uid];
    $form['status'] = ['#type' => 'value', '#value' => ApproverInterface::APPROVAL_PENDING];
    $form['created'] = ['#type' => 'value', '#value' => $request_time];
    $form['changed'] = ['#type' => 'value', '#value' => $request_time];

    // Ajax-effect to add a new approver.
    if ($form['step']['#value'] < 1) {
      $form['actions']['add'] = [
        '#type' => 'submit',
        '#value' => $this->t('Add approval'),
        '#submit' => ['::addSubmit'],
        '#ajax' => [
          'wrapper' => 'add-approval-wrapper',
          'callback' => '::prompt',
        ],
      ];
    }
    // This is the actual approver form.
    else {
      $form['subform']['approver_uid'] = [
        '#type' => 'entity_autocomplete',
        '#target_type' => 'user',
        '#selection_handler' => 'workbench_approver',
        '#selection_settings' => [
          'match_operator' => 'CONTAINS',
          'filters' => [
            'content_entity_id' => $nid,
            'content_entity_revision_id' => $vid,
            'content_entity_type_id' => 'node',
            'workflow_id' => $workflow_id,
          ],
        ],
        '#tags' => FALSE,
        '#title' => $this->t('User name'),
        '#placeholder' => $this->t('Type a name to search'),
        '#default_value' => $form_state->getValue(['subform', 'approver_uid']) ?? '',
        '#required' => TRUE,
      ];

      // Default deadline is one week.
      // @todo Make this configurable.
      $default_deadline = DrupalDateTime::createFromTimestamp($request_time + (7 * 24 * 60 * 60));

      $form['subform']['deadline'] = [
        '#type' => 'datetime',
        '#title' => $this->t('Due date'),
        '#default_value' => $form_state->getValue(['subform', 'deadline']) ?? $default_deadline,
        '#required' => TRUE,
      ];
      $form['subform']['required'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Approval from this user is required'),
        '#default_value' => $form_state->getValue(['subform', 'required']) ?? '',
      ];

      $form['subform']['notes'] = [
        '#type' => 'textarea',
        '#rows' => 3,
        '#cols' => 40,
        '#resizable' => 'both',
        '#title' => $this->t('Notes'),
        '#placeholder' => $this->t('Add any notes for the approver.'),
        '#default_value' => $form_state->getValue(['subform', 'notes']) ?? '',
        '#description' => $this->t('Notes are optional and may not contain HTML.'),
      ];

      // This button will save the page and refresh.
      $form['actions']['submit'] = [
        '#type' => 'submit',
        '#value' => $this->t('Request approval'),
        '#attributes' => ['class' => ['button--primary']],
      ];

      // On cancel, no validation of the form elements.
      // Go back to initial form state.
      $limit_validation_errors = [
        ['subform', 'approver_uid'],
        ['subform', 'deadline'],
        ['subform', 'required'],
        ['subform', 'notes'],
      ];
      $form['actions']['cancel'] = [
        '#type' => 'submit',
        '#value' => $this->t('Cancel'),
        '#limit_validation_errors' => $limit_validation_errors,
        '#submit' => ['::cancelAdd'],
        '#ajax' => [
          'wrapper' => 'add-approval-wrapper',
          'callback' => '::prompt',
        ],
      ];
    }

    $form['#prefix'] = '<div id="add-approval-wrapper">';
    $form['#suffix'] = '</div>';

    return $form;
  }

  /**
   * Wizard callback function.
   *
   * @param array $form
   *   Form API form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form API form.
   *
   * @return array
   *   Form array.
   */
  public function prompt(array $form, FormStateInterface $form_state) {
    return $form;
  }

  /**
   * Ajax callback that moves the form to the next step and rebuilds the form.
   *
   * @param array $form
   *   The Form API form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The FormState object.
   *
   * @return array
   *   The Form API form.
   */
  public function addSubmit(array $form, FormStateInterface $form_state) {
    $form_state->setValue('step', 1);
    $form_state->setRebuild();
    return $form;
  }

  /**
   * Ajax callback that cancels the process.
   *
   * @param array $form
   *   The Form API form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The FormState object.
   *
   * @return array
   *   The Form API form.
   */
  public function cancelAdd(array $form, FormStateInterface $form_state) {
    $form_state->setValue('step', 0);
    $form_state->setRebuild();
    return $form;
  }

  /**
   * Save away the current information.
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $values = [
      'workflow_id',
      'moderation_state',
      'content_entity_type_id',
      'content_entity_id',
      'content_entity_revision_id',
      'uid',
      'status',
      'required',
      'approver_uid',
      'deadline',
      'notes',
      'created',
      'changed',
      'langcode',
    ];

    $new_approver = [];

    foreach ($values as $value) {
      $new_approver[$value] = $form_state->getValue($value);
    }

    // Convert date to timestamp.
    $new_approver['deadline'] = strtotime($new_approver['deadline']);

    $approver_storage = $this->entityTypeManager->getStorage('approver');

    // The create() method checks for existing revisions.
    $approver = $approver_storage->create($new_approver);

    $approver->save();

    $this->messenger()->addMessage('Approval request assigned.');
  }

}
