<?php

namespace Drupal\workbench_approver\Form;

use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;

/**
 * Base form for common actions.
 */
class ApproverFormBase extends ContentEntityForm {

  /**
   * The fields to display on the form.
   *
   * @var string[]
   */
  public $allowedFields = [];

  /**
   * The status to set for the approval request.
   *
   * @var int
   */
  public $status;

  /**
   * Indicates that the form should have a Cancel action.
   *
   * @var bool
   */
  public $canCancel = TRUE;

  /**
   * Get the allowed fields for this form instance.
   *
   * @return string[]
   *   An array of field names.
   */
  private function getAllowedFields(): array {
    return $this->allowedFields;
  }

  /**
   * Gets the new status for this form instance.
   *
   * @return int
   *   The status to set.
   */
  private function getStatus(): int {
    return $this->status;
  }

  /**
   * Gets the cancel setting for this form.
   *
   * @return bool
   *   True if the form should have a cancel button.
   */
  private function allowCancel(): bool {
    return $this->canCancel;
  }

  /**
   * Remove fields that should not be edited in this context.
   *
   * @param array $form
   *   The form being built.
   *
   * @return void
   *   The $form is updated by reference.
   */
  public function removeFields(array &$form) {
    $allowed = $this->getAllowedFields();

    foreach (Element::children($form) as $element) {
      if (!in_array($element, $allowed, TRUE)) {
        $form[$element]['#access'] = FALSE;
      }
    }

  }

  /**
   * Sets the new status of the record.
   *
   * @param array $form
   *   The form being built.
   *
   * @return void
   *   The $form is updated by reference.
   */
  public function setStatus(array &$form) {
    $form['status']['widget'][0]['value']['#default_value'] = $this->getStatus();
  }

  /**
   * Sets the notes field to empty.
   *
   * @param array $form
   *   The form being built.
   *
   * @return void
   *   The $form is updated by reference.
   */
  public function resetNotes(array &$form) {
    // Zero out the notes field.
    $form['notes']['widget'][0]['value']['#rows'] = 4;
    $form['notes']['widget'][0]['value']['#default_value'] = '';
    $form['notes']['widget'][0]['value']['#description'] = $this->t('Add any clarifying notes to accompany your review');
  }

  /**
   * Gets the title for the form submit button.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   A translatable title.
   */
  public function getSubmitTitle() {
    return $this->t('Submit');
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    // @todo This might move to the base class.
    $actions = [];
    $actions['submit'] = [
      '#type' => 'submit',
      '#value' => $this->getSubmitTitle(),
      '#submit' => ['::submitForm', '::save'],
    ];

    if ($this->allowCancel()) {
      $this->addCancelAction($actions);
    }

    return $actions;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Always save a new revision.
    $this->entity->setNewRevision();
    parent::submitForm($form, $form_state);
  }

  /**
   * Adds a default cancel button to the form.
   *
   * @param array $actions
   *   The actions array.
   *
   * @return void
   *   The array is updated by reference.
   */
  public function addCancelAction(array &$actions) {
    // @todo Make this ajax-sensitive.
    $actions['cancel'] = [
      '#type' => 'submit',
      '#value' => $this->t('Cancel'),
      '#submit' => ['::cancelForm'],
    ];
  }

  /**
   * Submit handler for cancelling a form action.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return void
   *   No return value required.
   */
  public function cancelForm(array &$form, FormStateInterface $form_state) {
    // Do nothing and return.
    // @todo This works for AJAX, but does nothing for non-ajax versions.
  }

}
