<?php

namespace Drupal\workbench_approver\Services;

use Drupal\content_moderation\ModerationInformationInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\RevisionableStorageInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\workbench_approver\ApproverEvents;
use Drupal\workbench_approver\Entity\Approver;
use Drupal\workbench_approver\Entity\ApproverInterface;
use Drupal\workbench_approver\Event\GetPotentialApproversEvent;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

/**
 * Service to load potential approvers.
 *
 * This service is mainly used by the ApproverSelection
 * autocomplete class.
 *
 * We might be able to use the ApproverStorage service here,
 * but we want to make this explicitly swappable with a
 * service decorator.
 */
class ApproverManager implements ApproverManagerInterface {

  use StringTranslationTrait;

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected AccountProxyInterface $currentUser;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Contracts\EventDispatcher\EventDispatcherInterface
   */
  protected EventDispatcherInterface $eventDispatcher;

  /**
   * The moderation information service.
   *
   * @var \Drupal\content_moderation\ModerationInformationInterface
   */
  protected ModerationInformationInterface $moderationInfo;

  /**
   * The approval entity storage service.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $approverStorage;

  /**
   * Constructs a new ApproverManager object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity type manager.
   * @param \Drupal\content_moderation\ModerationInformationInterface $moderation_info
   *   The moderation information service.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The current user.
   * @param \Symfony\Contracts\EventDispatcher\EventDispatcherInterface $dispatcher
   *   The event dispatcher.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    ModerationInformationInterface $moderation_info,
    AccountProxyInterface $current_user,
    EventDispatcherInterface $dispatcher,
  ) {
    $this->entityTypeManager = $entityTypeManager;
    $this->approverStorage = $entityTypeManager->getStorage('approver');
    $this->moderationInfo = $moderation_info;
    $this->currentUser = $current_user;
    $this->eventDispatcher = $dispatcher;
  }

  /**
   * Getter method for approver storage variable.
   *
   * @return \Drupal\Core\Entity\EntityStorageInterface
   *   The approval entity storage service.
   */
  public function getApproverStorage(): EntityStorageInterface {
    return $this->approverStorage;
  }

  /**
   * {@inheritdoc}
   */
  public function isAssigner(): bool {
    return $this->currentUser->hasPermission('assign workbench approver');
  }

  /**
   * {@inheritdoc}
   */
  public function isViewer(): bool {
    return $this->currentUser->hasPermission('view workbench approver information');
  }

  /**
   * {@inheritdoc}
   */
  public function getAllowedRoles(): array {
    $rids = [];
    $role_storage = $this->entityTypeManager->getStorage('user_role');

    // Check that roles have our permission.
    // @todo In future, we might have permissions by content type.
    foreach ($role_storage->loadMultiple() as $role) {
      /** @var \Drupal\user\RoleInterface $role */
      if ($role->hasPermission('use workbench approver')) {
        $rids[] = $role->id();
      }
    }

    return $rids;
  }

  /**
   * {@inheritdoc}
   */
  public function getCurrentApproverIds(array $arguments): array {
    // Set the query.
    $query = $this->approverStorage->getAggregateQuery()->accessCheck(FALSE);

    // Loop through the conditions provided.
    foreach ($arguments as $arg => $value) {
      $query->condition($arg, $value);
    }

    $query->condition('status', ApproverInterface::APPROVAL_RESCINDED, 'NOT IN');

    // Get results.
    $results = $query->groupBy('approver_uid')->execute();

    return array_column($results, 'approver_uid');
  }

  /**
   * {@inheritdoc}
   */
  public function getPotentialApproverIds(array $arguments): array {
    // Set the user query.
    $user_storage = $this->entityTypeManager->getStorage('user');
    $query = $user_storage->getAggregateQuery()->accessCheck(FALSE);
    $query->condition('status', 1);

    // Filter by allowed role.
    $rids = $this->getAllowedRoles();
    if (count($rids) > 0) {
      $query->condition('roles', $rids, 'IN');
    }

    // Assigned users cannot be re-assigned.
    // However, for audit tracking, deleted users can be reassigned.
    $disallowed_uids = $this->getCurrentApproverIds($arguments);
    if (count($disallowed_uids) > 0) {
      $query->condition('uid', $disallowed_uids, 'NOT IN');
    }

    // Get results.
    $results = $query->groupBy('uid')->execute();
    $uids = array_column($results, 'uid');

    $event = new GetPotentialApproversEvent($uids, $arguments);
    $this->eventDispatcher->dispatch($event, ApproverEvents::WORKBENCH_APPROVER_GET_POTENTIAL_APPROVERS);

    return $event->getPotentialApproverIds();
  }

  /**
   * {@inheritdoc}
   */
  public function getBlockingApprovalRequests(ContentEntityInterface $entity, bool $get_latest_revision = TRUE): array {
    if ($get_latest_revision) {
      $entity = $this->getEntityRevision($entity);
    }
    $requests = [];
    foreach ($this->getApprovalRequests($entity) as $approver) {
      if ($approver->isBlocking()) {
        $requests[] = $approver;
      }
    }

    return $requests;
  }

  /**
   * {@inheritdoc}
   */
  public function getPendingApprovalRequests(ContentEntityInterface $entity): array {
    $requests = [];
    foreach ($this->getApprovalRequests($entity) as $approver) {
      if ($approver->isPending()) {
        $requests[] = $approver;
      }
    }

    return $requests;
  }

  /**
   * {@inheritdoc}
   */
  public function getApprovalRequestByUser(ContentEntityInterface $entity, AccountInterface $account, $active = TRUE): ?ApproverInterface {
    $request = NULL;
    $params = [
      'approver_uid' => $account->id(),
    ];
    foreach ($this->getApprovalRequests($entity, $params) as $approver) {
      // For some reason, these do not type match.
      if (!$active || $approver->isPending()) {
        $request = $approver;
      }
    }

    return $request;
  }

  /**
   * {@inheritdoc}
   */
  public function isBlocked(ContentEntityInterface $entity): bool {
    if ($entity->isNew()) {
      return FALSE;
    }

    $active = FALSE;
    foreach ($this->getApprovalRequests($entity) as $approver) {
      if ($approver->isBlocking()) {
        $active = TRUE;
      }
    }

    return $active;
  }

  /**
   * {@inheritdoc}
   */
  public function getApprovalRequests(ContentEntityInterface $entity, array $additional_parameters = []): array {
    $workflow = $this->moderationInfo->getWorkflowForEntity($entity);
    $list = [];
    // Entities not under moderation will have a NULL workflow. Entities in
    // preview will not have a value for revision id, which will
    // result in callstack errors if attempts are made to get the revision id.
    // @phpstan-ignore-next-line
    if ($workflow !== NULL && $entity->in_preview !== TRUE) {
      $parameters = [
        'content_entity_type_id' => $entity->getEntityTypeId(),
        'content_entity_id' => $entity->id(),
        'content_entity_revision_id' => $entity->getRevisionId(),
        'workflow_id' => $workflow->id(),
        'langcode' => $entity->language()->getId(),
      ];

      $parameters += $additional_parameters;
      $approvers = $this->approverStorage->loadByProperties($parameters);

      // We always filter out "rescinded" elements.
      // @todo This may be better as an entity query, not loadByProperties.
      /** @var \Drupal\workbench_approver\Entity\ApproverInterface $approver */
      foreach ($approvers as $approver) {
        if ($approver->getStatus() !== ApproverInterface::APPROVAL_RESCINDED) {
          $list[] = $approver;
        }
      }
    }

    return $list;
  }

  /**
   * {@inheritdoc}
   */
  public function getRevisionUrl(ContentEntityInterface $entity): Url {
    $rel = $entity->getEntityType()->hasLinkTemplate('revision') ? 'revision' : 'canonical';
    if ($rel === 'revision' && !$entity->isDefaultRevision()
                            && $entity->isLatestTranslationAffectedRevision()
                            && $entity->hasLinkTemplate('latest-version')) {
      $rel = 'latest-version';
    }
    return $entity->toUrl($rel);
  }

  /**
   * {@inheritdoc}
   */
  public function entityUpdate(ContentEntityInterface $entity): void {
    if ($this->moderationInfo->isModeratedEntity($entity)) {
      // Pick up values from hook_entity_presave().
      // @phpstan-ignore-next-line
      $is_latest_revision = $entity->__get('workbench_approver_status') ?? FALSE;

      // Get active approval requests for the current entity.
      // The saved entity records its parent in getLoadedRevisionId(),
      // which is "the loaded revision ID before the new revision was set."
      // However, on revert, we are not dealing with the latest revision.
      // @phpstan-ignore-next-line
      $latest_revision_id = $entity->__get('workbench_approver_latest_revision_id') ?? $entity->getLoadedRevisionId();

      $parameters = [
        'content_entity_type_id' => $entity->getEntityTypeId(),
        'content_entity_id' => $entity->id(),
        'content_entity_revision_id' => $latest_revision_id,
      ];
      /** @var \Drupal\workbench_approver\Entity\ApproverInterface[] $approvals */
      $approvals = $this->approverStorage->loadByProperties($parameters);

      $state_changed = $this->moderationInfo->getOriginalState($entity)->id() !== $entity->get('moderation_state')->getString();

      // Act on the latest revision to reassign requests.
      // Create new requests if the state did not change.
      $options = [];
      if (!$state_changed && $is_latest_revision) {
        $options['allow_creation_status'] = [
          ApproverInterface::APPROVAL_PENDING,
          ApproverInterface::APPROVAL_REJECTED,
        ];
        $this->createNewApprovalRequests($approvals, $entity, $options);
      }
      // Reassign approval requests as applicable.
      $this->reassignCurrentApprovalRequests($approvals, $entity, $state_changed);
    }
  }

  /**
   * Creates new requests if the current version changes.
   *
   * @param \Drupal\workbench_approver\Entity\ApproverInterface[] $approvals
   *   The approval requests for this revision.
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity being updated.
   * @param array $options
   *   Array containing the conditions under the approval has to be created.
   *
   * @return void
   *   No return value.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function createNewApprovalRequests(array $approvals, ContentEntityInterface $entity, array $options = []): void {
    foreach ($approvals as $approval) {
      if (count($options) === 0 || (isset($options['allow_creation_status']) && in_array($approval->getStatus(), $options['allow_creation_status'], TRUE))) {
        // A revision reassigns active approval requests to the new revision.
        // This action should create a new approval request.
        $this->createNewApprovalRequest($entity, $approval);
      }
    }
  }

  /**
   * Creates a new approval request after entity revision.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity being updated.
   * @param \Drupal\workbench_approver\Entity\ApproverInterface $approval
   *   An existing approval for the current revision.
   *
   * @return void
   *   No return value.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function createNewApprovalRequest(ContentEntityInterface $entity, ApproverInterface $approval): void {
    /** @var \Drupal\node\Entity\Node $entity */
    $workflow = $this->moderationInfo->getWorkflowForEntity($entity);
    $values = [
      'workflow_id' => $workflow->id(),
      'moderation_state' => $entity->get('moderation_state')->getString(),
      'content_entity_type_id' => $entity->getEntityTypeId(),
      'content_entity_id' => $entity->id(),
      'content_entity_revision_id' => $entity->getRevisionId(),
      'uid' => $approval->getOwnerId(),
      'status' => ApproverInterface::APPROVAL_PENDING,
      'required' => $approval->get('required')->getString(),
      'approver_uid' => $approval->getApproverId(),
      'deadline' => $approval->get('deadline')->getString(),
      'notes' => $approval->get('notes')->getString(),
      'created' => $entity->getChangedTime(),
      'changed' => $entity->getChangedTime(),
      'langcode' => $entity->language()->getId(),
    ];

    $this->approverStorage->create($values)->save();
  }

  /**
   * Changes the status on requests if the current version changes.
   *
   * @param \Drupal\workbench_approver\Entity\ApproverInterface[] $approvals
   *   The approval requests for this revision.
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity being updated.
   * @param bool $state_changed
   *   Indicates that the revision changed moderation state.
   *
   * @return void
   *   No return value.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function reassignCurrentApprovalRequests(array $approvals, ContentEntityInterface $entity, bool $state_changed): void {
    foreach ($approvals as $approval) {
      // This action should set a new revision of the request.
      // Note that we do not reset the original owner of the entity.
      $status = $this->getNewStatus($approval);
      $approval->set('status', $status);
      // @phpstan-ignore-next-line
      $approval->set('changed', $entity->getChangedTime());
      if ($state_changed) {
        $approval->set('notes', $this->t('The revision has been approved.'));
      }
      else {
        $approval->set('notes', $this->t('A new revision was created, and a new approval is required.'));
      }
      // @todo Revisions might become optional.
      $approval->setNewRevision(TRUE);
      $approval->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function modifyApprovalRequestsOfBlockedApprover(array $approvals): void {
    // @todo We may be able to refactor these three methods to reduce code duplication.
    /** @var \Drupal\workbench_approver\Entity\Approver $approval */
    foreach ($approvals as $approval) {
      // Get the parent entity of the approval.
      $entity_revision = $this->getParentEntityRevision($approval);

      // Change active approvals to 'RESCINDED' and leave message in the notes
      // of all approvals having the user blocked as the approver.
      if ($approval->isActive()) {
        $note = $this->t('The status of this request has been changed to "%status_name". The approver of this request with username "%username" has been blocked.', [
          '%status_name' => Approver::statusName(ApproverInterface::APPROVAL_RESCINDED),
          '%username' => $approval->getApproverLabel(),
        ]);
        $approval->set('notes', $note);
        $approval->set('status', ApproverInterface::APPROVAL_RESCINDED);
        // @phpstan-ignore-next-line
        $approval->set('changed', $entity_revision->getChangedTime());
        $approval->setNewRevision(TRUE);
        $approval->save();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function modifyApprovalRequestsOfDeletedApprover(array $approvals): void {
    // @todo We may be able to refactor these three methods to reduce code duplication.
    /** @var \Drupal\workbench_approver\Entity\Approver $approval */
    foreach ($approvals as $approval) {
      // Get the parent entity of the approval.
      $entity_revision = $this->getParentEntityRevision($approval);

      // If the deleted approver is also the actual owner, delete the request
      // approval. This is to avoid leaving a reference to the deleted user as
      // the approver of this request. Neither make sense to change both to
      // Anonymous (user 0), so deleting the request in this case makes sense.
      if ($approval->getApproverId() == $approval->getOwnerId()) {
        $approval->delete();
        continue;
      }

      // Change active approvals to 'RESCINDED', reassign to the Owner and
      // leave a message in the notes of all approvals having the deleted user
      // as the approver.
      if ($approval->isActive()) {
        $approval->set('status', ApproverInterface::APPROVAL_RESCINDED);
      }
      $note = $this->t('The status of this request has been changed to "%status_name". The approver of this request with username "%username" has been deleted.', [
        '%status_name' => Approver::statusName(ApproverInterface::APPROVAL_RESCINDED),
        '%username' => $approval->getApproverLabel(),
      ]);
      $approval->set('notes', $note);
      // @phpstan-ignore-next-line
      $approval->set('changed', $entity_revision->getChangedTime());
      $approval->set('approver_uid', $approval->getOwnerId());
      $approval->setNewRevision(TRUE);
      $approval->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function modifyApprovalRequestsOfDeletedOwner(array $approvals): void {
    // @todo We may be able to refactor these three methods to reduce code duplication.
    foreach ($approvals as $approval) {
      assert($approval instanceof ApproverInterface);
      // Get the parent entity of the approval.
      $entity_revision = $this->getParentEntityRevision($approval);

      // If the deleted user is the owner of an approval request, reassign the
      // approvals to Anonymous user (0) and leave message on notes.
      $note = $this->t('The owner of this request with username "%username" has been deleted. The new owner of this approval is now the "Anonymous" user (0).', [
        '%username' => $approval->getOwnerLabel(),
      ]);
      $approval->set('notes', $note);
      // @phpstan-ignore-next-line
      $approval->set('changed', $entity_revision->getChangedTime());
      $approval->set('uid', 0);
      $approval->setNewRevision(TRUE);
      $approval->save();
    }
  }

  /**
   * Given an approval returns the corresponding parent entity revision.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getParentEntityRevision($approval): ContentEntityInterface {
    // Get the parent entity of the approval.
    $node_storage = $this->entityTypeManager->getStorage('node');
    /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
    $entity = $node_storage->load($approval->getContentEntityId());
    return $this->getEntityRevision($entity, $approval->getContentEntityRevisionId());
  }

  /**
   * Negotiates status update logic.
   *
   * @param \Drupal\workbench_approver\Entity\ApproverInterface $approval
   *   The approval request being updated.
   *
   * @return int
   *   Return a new status.
   */
  private function getNewStatus(ApproverInterface $approval): int {
    $status = $approval->getStatus();

    // Pending approvals are marked as outdated when node state changed.
    if ($status === ApproverInterface::APPROVAL_PENDING) {
      $status = ApproverInterface::APPROVAL_OUTDATED;
    }

    return $status;
  }

  /**
   * {@inheritdoc}
   */
  public function entityDelete(ContentEntityInterface $entity): void {
    // Deleting an entity deletes all translations.
    $parameters = [
      'content_entity_type_id' => $entity->getEntityTypeId(),
      'content_entity_id' => $entity->id(),
    ];

    $this->deleteApprovalRequests($parameters);
  }

  /**
   * {@inheritdoc}
   */
  public function entityRevisionDelete(ContentEntityInterface $entity): void {
    // Deleting an entity revision deletes all translations.
    $parameters = [
      'content_entity_type_id' => $entity->getEntityTypeId(),
      'content_entity_id' => $entity->id(),
      'content_entity_revision_id' => $entity->getRevisionId(),
    ];

    $this->deleteApprovalRequests($parameters);
  }

  /**
   * {@inheritdoc}
   */
  public function entityTranslationDelete(ContentEntityInterface $entity): void {
    // Translation deletion is a separate hook.
    // @todo We do not have test coverage for this case.
    $parameters = [
      'content_entity_type_id' => $entity->getEntityTypeId(),
      'content_entity_id' => $entity->id(),
      'langcode' => $entity->language()->getId(),
    ];

    $this->deleteApprovalRequests($parameters);
  }

  /**
   * Deletes current approval requests, based on query conditions.
   *
   * This method is internal and not part of the interface.
   *
   * @param array $parameters
   *   An array of arguments, taken from the Approver entity type.
   *
   * @return void
   *   No return value.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function deleteApprovalRequests(array $parameters): void {
    foreach ($this->approverStorage->loadByProperties($parameters) as $approver) {
      $approver->delete();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityRevision(ContentEntityInterface $entity, int $revision_id = NULL): ?ContentEntityInterface {
    $node_storage = $this->entityTypeManager->getStorage($entity->getEntityTypeId());
    assert($node_storage instanceof RevisionableStorageInterface);

    if (is_null($revision_id)) {
      $revision_id = $node_storage->getLatestRevisionId($entity->id());
    }

    $revision = $node_storage->loadRevision($revision_id);

    return $revision;
  }

  /**
   * {@inheritdoc}
   */
  public function getApprovalRequestsForLatestRevision(ContentEntityInterface $entity): array {
    $latest_revision = $this->getEntityRevision($entity);

    return $this->getApprovalRequests($latest_revision);
  }

  /**
   * {@inheritdoc}
   */
  public function getApprovalRequestIdsByEntityKeys(string $entity_key_id, int $entity_key_value): array {
    // In the entity keys for Approver entity 'owner' = 'uid'.
    if ($entity_key_id === 'owner') {
      $entity_key_id = 'uid';
    }

    // Set the user query.
    $query = $this->approverStorage->getQuery()
      ->accessCheck(FALSE)
      ->condition($entity_key_id, $entity_key_value);
    return $query->execute();
  }

}
