<?php

namespace Drupal\workbench_approver\Services;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\workbench_approver\Entity\ApproverInterface;

/**
 * Interface for the ApproverManager service.
 *
 * This service handles relationships between approval request
 * entities and other systems. Methods related entirely to
 * approval requests are in the Entity class or its Storage.
 */
interface ApproverManagerInterface {

  /**
   * Getter method for approver storage variable.
   *
   * @return \Drupal\Core\Entity\EntityStorageInterface
   *   The approval entity storage service.
   */
  public function getApproverStorage(): EntityStorageInterface;

  /**
   * Determine the status of the user for the content view.
   *
   * @return bool
   *   Whether the current user is an assigner.
   */
  public function isAssigner(): bool;

  /**
   * Determine the status of the user for the content view.
   *
   * @return bool
   *   Whether the current user is a viewer.
   */
  public function isViewer(): bool;

  /**
   * Determines if a revision has blocking approvals.
   *
   * A blocking approval request is defined as:
   *
   * - ApproverInterface::isRequired() === TRUE
   *
   * AND one of the following:
   * - ApproverInterface::getStatus() === ApproverInterface::APPROVAL_PENDING
   * - ApproverInterface::getStatus() === ApproverInterface::APPROVAL_REJECTED
   *
   * All other states are for auditing.
   *
   * If there are no open approval requests that meet this condition,
   * then the content can be transitioned to a new state.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The parent entity revision.
   *
   * @return bool
   *   TRUE if at least one required reviewer has not approved. FALSE otherwise.
   */
  public function isBlocked(ContentEntityInterface $entity): bool;

  /**
   * Gets approval requests for the loaded revision.
   *
   * Note that this method responds to the loaded revision
   * of the entity. If you need to ensure you are looking
   * at the approval requests on the latest revision, use
   * getApprovalRequestsForLatestRevision() instead.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The parent entity revision.
   * @param array $additional_parameters
   *   Any additional properties to pass to the query.
   *
   * @return \Drupal\workbench_approver\Entity\ApproverInterface[]
   *   An array of approval requests.
   */
  public function getApprovalRequests(ContentEntityInterface $entity, array $additional_parameters = []): array;

  /**
   * Get all users with the 'use workbench approver' permission.
   *
   * @return array
   *   An array of role ids.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getAllowedRoles(): array;

  /**
   * Gets the user ids of approvers, based on query conditions.
   *
   * @param array $arguments
   *   An array of arguments, taken from the Approver entity type.
   *   Default keys are:
   *    'content_entity_id' -- the entity id
   *    'content_entity_revision_id', -- the revision id
   *    'content_entity_type_id' -- the entity type id (node)
   *    'workflow' -- the workflow id
   *
   *   See AddApproverForm. This API is still under development.
   *   It is designed for flexibility when overriding the form
   *   or creating a custom service implementation.
   *
   *   Note that this function should be called within the
   *   getPotentialApproverIds() method in order to filter out
   *   users that are already assigned.
   *
   * @return array
   *   An array of user ids.
   */
  public function getCurrentApproverIds(array $arguments): array;

  /**
   * Gets the user ids of approvers, based on query conditions.
   *
   * This list will explicitly filter out current approvers.
   *
   * @param array $arguments
   *   An array of arguments, taken from the Approver entity type.
   *   Default keys are:
   *    'content_entity_id' -- the entity id
   *    'content_entity_revision_id', -- the revision id
   *    'content_entity_type_id' -- the entity type id (node)
   *    'workflow' -- the workflow id.
   *
   * @return array
   *   An array of user ids.
   */
  public function getPotentialApproverIds(array $arguments): array;

  /**
   * Gets all blocking assignments for an entity.
   *
   * By default, this method will load the latest revision before
   * checking the approval requests.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity under inspection, which may be the latest revision.
   * @param bool $get_latest_revision
   *   Flag to ensure we load the latest revision.
   *
   * @return \Drupal\workbench_approver\Entity\ApproverInterface[]
   *   An array of approval requests.
   */
  public function getBlockingApprovalRequests(ContentEntityInterface $entity, bool $get_latest_revision = TRUE): array;

  /**
   * Gets all pending assignments for an entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity revision under inspection.
   *
   * @return \Drupal\workbench_approver\Entity\ApproverInterface[]
   *   An array of approval requests.
   */
  public function getPendingApprovalRequests(ContentEntityInterface $entity): array;

  /**
   * Gets the assignments for a user assigned to an entity.
   *
   * Note that only one assignment should be allowed per revision.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity revision under inspection.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   Assignments will only include those for the specified user.
   * @param bool $active
   *   Flag to only return active records.
   *
   * @return \Drupal\workbench_approver\Entity\ApproverInterface|null
   *   The approval request or NULL.
   */
  public function getApprovalRequestByUser(ContentEntityInterface $entity, AccountInterface $account, bool $active = TRUE): ?ApproverInterface;

  /**
   * Gets the proper revision URL for an entity.
   *
   * Entities with a published revision use `latest-version`.
   *
   * See https://www.drupal.org/project/drupal/issues/2906455.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity revision under inspection.
   *
   * @return \Drupal\Core\Url
   *   The Url object for the proper revision link.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function getRevisionUrl(ContentEntityInterface $entity): Url;

  /**
   * Responds to entity revision creation (hook_entity_update).
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The updated entity.
   *
   * @return void
   *   No return value.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function entityUpdate(ContentEntityInterface $entity): void;

  /**
   * Responds to entity deletion (hook_entity_delete).
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity being deleted.
   *
   * @return void
   *   No return value.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function entityDelete(ContentEntityInterface $entity): void;

  /**
   * Responds to entity deletion (hook_entity_revision_delete).
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity revision being deleted.
   *
   * @return void
   *   No return value.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function entityRevisionDelete(ContentEntityInterface $entity): void;

  /**
   * Responds to translation deletion (hook_entity_translation_delete).
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity translation being deleted.
   *
   * @return void
   *   No return value.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function entityTranslationDelete(ContentEntityInterface $entity): void;

  /**
   * Returns a revision of an entity.
   *
   * If no revision id is given the latest revision will be returned.
   *
   * This method is helpful because getApprovalRequests() is based on the
   * entity revision passed to it. If we need to get approval requests
   * for the latest revision while looking at a different revision,
   * then we need to load the latest revision.
   *
   * See workbench_approver_form_node_revision_revert_confirm_alter().
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity under inspection.
   * @param int|null $revision_id
   *   The revision id to load.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface
   *   The latest entity revision.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getEntityRevision(ContentEntityInterface $entity, int $revision_id = NULL): ?ContentEntityInterface;

  /**
   * Gets approval requests for the latest revision.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity under inspection.
   *
   * @return \Drupal\workbench_approver\Entity\ApproverInterface[]
   *   An array of approval requests.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getApprovalRequestsForLatestRevision(ContentEntityInterface $entity): array;

  /**
   * Gets all the approval requests for the given user as the approver.
   *
   * @param string $entity_key_id
   *   The label of the entity key as defined in the "Approver" entity.
   * @param int $entity_key_value
   *   The value to use for the entity hey (for example the Approver Id or the
   *   Owner Id).
   *
   * @return array
   *   The array of Approvals Ids.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  public function getApprovalRequestIdsByEntityKeys(string $entity_key_id, int $entity_key_value): array;

  /**
   * When an approver is blocked actions need to be done on the approvals.
   *
   * @param array $approvals
   *   The request approvals to modify.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function modifyApprovalRequestsOfBlockedApprover(array $approvals): void;

  /**
   * When an approver is deleted actions need to be done on the approvals.
   *
   * @param array $approvals
   *   The request approvals to modify.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function modifyApprovalRequestsOfDeletedApprover(array $approvals): void;

  /**
   * When an owner is deleted actions need to be done on the approvals.
   *
   * @param array $approvals
   *   The request approvals to modify.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function modifyApprovalRequestsOfDeletedOwner(array $approvals): void;

}
