<?php

namespace Drupal\Tests\workbench_approver\Functional;

use Drupal\Core\Session\AccountInterface;
use Drupal\node\NodeInterface;
use Drupal\workbench_approver\Entity\ApproverInterface;

/**
 * Tests content on the approvals tab.
 *
 * @group workbench_approver
 */
class ApproverControllerTest extends WorkbenchApproverTestBase {

  /**
   * A node with active assignments.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $assignedNode;

  /**
   * A node with prior and active assignments.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $priorNode;

  /**
   * A node without assignments.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $plainNode;

  /**
   * A node not within a workflow.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $notAllowedNode;

  /**
   * A published node with assignments.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $publishedNode;

  /**
   * A node without a valid transition options.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $archivedNode;

  /**
   * An array of test nodes.
   *
   * @var array
   */
  protected $data;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->assignedNode = $this->createNodeForTest('Assigned node');
    $this->addApprovalRequest($this->adminUser, $this->assignedNode, ['required' => TRUE, 'notes' => 'Test note']);
    $this->addApprovalRequest($this->editorUser, $this->assignedNode, ['required' => FALSE]);

    $prior_node = $this->createNodeForTest('Prior node');
    $this->addApprovalRequest($this->adminUser, $prior_node, ['required' => FALSE]);
    $this->addApprovalRequest($this->editorUser, $prior_node, ['required' => FALSE]);
    $this->priorNode = $this->createRevision($prior_node);
    $this->addApprovalRequest($this->adminUser, $this->priorNode, ['required' => FALSE]);
    $this->addApprovalRequest($this->editorUser, $this->priorNode, ['required' => FALSE]);

    $this->plainNode = $this->drupalCreateNode([
      'type' => 'approved',
      'title' => 'Plain node',
      'moderation_state' => 'draft',
    ]);

    $this->notAllowedNode = $this->drupalCreateNode([
      'type' => 'non_approved',
      'title' => 'Non-approved node',
    ]);

    $published_node = $this->createNodeForTest('Published node');
    $this->addApprovalRequest($this->adminUser, $published_node, ['required' => FALSE]);
    $this->addApprovalRequest($this->editorUser, $published_node, ['required' => FALSE]);
    $this->publishedNode = $this->createRevision($published_node, 'published');

    $this->archivedNode = $this->drupalCreateNode([
      'type' => 'approved',
      'title' => 'Archived node',
      'moderation_state' => 'archived',
    ]);

    $this->data = [
      $this->assignedNode->id() => [
        'node' => $this->assignedNode,
        'active' => TRUE,
        'prior' => FALSE,
        'invalid' => FALSE,
        'published' => FALSE,
        'allowed' => TRUE,
      ],
      $this->priorNode->id() => [
        'node' => $this->priorNode,
        'active' => TRUE,
        'prior' => TRUE,
        'invalid' => FALSE,
        'published' => FALSE,
        'allowed' => TRUE,
      ],
      $this->plainNode->id() => [
        'node' => $this->plainNode,
        'active' => FALSE,
        'prior' => FALSE,
        'invalid' => FALSE,
        'published' => FALSE,
        'allowed' => TRUE,
      ],
      $this->publishedNode->id() => [
        'node' => $this->publishedNode,
        'active' => TRUE,
        'prior' => TRUE,
        'invalid' => FALSE,
        'published' => TRUE,
        'allowed' => TRUE,
      ],
      $this->archivedNode->id() => [
        'node' => $this->archivedNode,
        'active' => TRUE,
        'prior' => FALSE,
        'invalid' => TRUE,
        'published' => FALSE,
        'allowed' => TRUE,
      ],
      $this->notAllowedNode->id() => [
        'node' => $this->notAllowedNode,
        'active' => FALSE,
        'prior' => FALSE,
        'invalid' => FALSE,
        'published' => FALSE,
        'allowed' => FALSE,
      ],
    ];
  }

  /**
   * Tests the page as an observer.
   *
   * Users who can only view cannot see any revision data.
   */
  public function testObserver() {
    $this->drupalLogin($this->observerUser);

    foreach ($this->data as $data) {
      $this->assertApprovalPage($data);

      if ($data['allowed']) {
        $this->assertSession()->pageTextContains('You may view the approvals for this content.');
        // Run default tests.
        $this->assertNoSummaryText($data);
        $this->assertActiveContent($data);
        $this->assertPriorContent($data);
        $this->assertPublishedContent($data);
        $this->assertWorkflowText($data);
        $this->assertInvalidContent($data);
        $this->assertAddApproval($data);
        $this->assertOperationsLinks($data, $this->observerUser);
      }
    }
  }

  /**
   * Tests the page as an editor.
   */
  public function testEditor() {
    $this->drupalLogin($this->editorUser);

    foreach ($this->data as $data) {
      $this->assertApprovalPage($data);

      if ($data['allowed']) {
        // Run default tests.
        $this->assertSummaryText($data);
        $this->assertActiveContent($data);
        $this->assertPriorContent($data);
        $this->assertPublishedContent($data);
        $this->assertWorkflowText($data);
        $this->assertInvalidContent($data);
        $this->assertReviewRequested($data);
        $this->assertAddApproval($data);
        $this->assertOperationsLinks($data, $this->editorUser, TRUE);
      }
    }
  }

  /**
   * Tests the page as an administrator.
   */
  public function testAdministrator() {
    $this->drupalLogin($this->adminUser);

    foreach ($this->data as $data) {
      $this->assertApprovalPage($data);

      if ($data['allowed']) {
        $this->assertSummaryText($data, TRUE);
        $this->assertActiveContent($data);
        $this->assertPriorContent($data);
        $this->assertPublishedContent($data);
        $this->assertWorkflowText($data);
        $this->assertInvalidContent($data);
        // This user is an admin and an assigned approver.
        $this->assertReviewRequested($data, TRUE);
        $this->assertAddApproval($data, TRUE);
        $this->assertOperationsLinks($data, $this->adminUser, TRUE, TRUE);
      }
    }
  }

  /**
   * Tests the page as a workflow administrator.
   */
  public function testWorkflowAdministrator() {
    $this->drupalLogin($this->workflowAdminUser);

    foreach ($this->data as $data) {
      $this->assertApprovalPage($data);

      if ($data['allowed']) {
        $this->assertSummaryText($data, TRUE);
        $this->assertActiveContent($data);
        $this->assertPriorContent($data);
        $this->assertPublishedContent($data);
        $this->assertWorkflowText($data, TRUE);
        $this->assertInvalidContent($data, TRUE);
        // This user is an admin but not an assigned approver.
        $this->assertReviewRequested($data, TRUE, FALSE);
        $this->assertAddApproval($data, TRUE);
        $this->assertOperationsLinks($data, $this->workflowAdminUser, FALSE, TRUE);
        $this->assertActiveView($data);
        $this->assertPriorView($data);
      }
    }
  }

  /**
   * Creates a simple node for testing views.
   *
   * @param string $title
   *   The node title.
   *
   * @return \Drupal\node\NodeInterface
   *   The created node.
   */
  private function createNodeForTest($title): NodeInterface {
    $values = [
      'type' => 'approved',
      'title' => $title,
      'moderation_state' => 'published',
    ];
    $approved_node = $this->drupalCreateNode($values);
    $revised_node = $this->createRevision($approved_node);

    return $revised_node;
  }

  /**
   * Base page load test.
   *
   * @param array $data
   *   The node data for the test.
   */
  private function assertApprovalPage(array $data) {
    $node = $data['node'];
    $path = '/node/' . $node->id() . '/approver';
    $this->drupalGet($path);

    if ($data['allowed']) {
      $this->assertSession()->statusCodeEquals(200);
      $this->assertSession()->pageTextContains(sprintf('Approvals for %s', $node->label()));
    }
    else {
      $this->assertSession()->statusCodeEquals(403);
    }
  }

  /**
   * Asserts the default content summary is not present.
   *
   * @param array $data
   *   The node data for the test.
   */
  private function assertNoSummaryText(array $data) {
    $node = $data['node'];
    // No summary text.
    $this->assertSession()->pageTextNotContains('Current revision state:');
    $this->assertSession()->pageTextNotContains('Current revision id:');
    $this->assertSession()->pageTextNotContains('Current revision:');
    $link = $this->approverManager->getRevisionUrl($node)->toString();
    $this->assertSession()->linkByHrefNotExists($link);
  }

  /**
   * Asserts the default content summary.
   *
   * @param array $data
   *   The node data for the test.
   * @param bool $is_admin
   *   Indicates the user is an administrator.
   */
  private function assertSummaryText(array $data, bool $is_admin = FALSE) {
    $node = $data['node'];
    // @todo Prior approvals are not appearing properly.
    // Remove the published check when they are.
    $no_summary = [
      $this->plainNode->id(),
      $this->publishedNode->id(),
      $this->archivedNode->id(),
    ];
    if (!$is_admin && in_array($node->id(), $no_summary, TRUE)) {
      $this->assertNoSummaryText($data);
    }
    else {
      $this->assertSession()->pageTextContains('Current revision state: ' . $node->get('moderation_state')->getString());
      $this->assertSession()->pageTextContains('Current revision id: ' . $node->getRevisionId());
      $this->assertSession()->pageTextContains('Current revision: ' . $node->label());
      $this->assertSession()->linkExists($node->label(), 0);
      // Expected link format.
      $link = $this->approverManager->getRevisionUrl($node)->toString();
      $this->assertSession()->linkByHrefExists($link);
    }
  }

  /**
   * Asserts the default active content view.
   *
   * @param array $data
   *   The node data for the test.
   */
  private function assertActiveContent(array $data) {
    if ($data['active']) {
      $this->assertSession()->pageTextNotContains('There are no active approval requests for the current revision.');
    }
    else {
      $this->assertSession()->pageTextContains('There are no active approval requests for the current revision.');
    }
  }

  /**
   * Asserts the default prior content view.
   *
   * @param array $data
   *   The node data for the test.
   */
  private function assertPriorContent(array $data) {
    if ($data['prior']) {
      $this->assertSession()->pageTextNotContains('There are no prior approval requests for this content.');
    }
    else {
      $this->assertSession()->pageTextContains('There are no prior approval requests for this content.');
    }
  }

  /**
   * Asserts the default invalid transition message.
   *
   * @param array $data
   *   The node data for the test.
   * @param bool $is_workflow_admin
   *   Indicates the user can administer workflows.
   */
  private function assertInvalidContent(array $data, bool $is_workflow_admin = FALSE) {
    if ($data['invalid'] && !$is_workflow_admin) {
      $this->assertSession()->pageTextContains('No current approvals are required for this content.');
    }
    else {
      $this->assertSession()->pageTextNotContains('No current approvals are required for this content.');
    }
  }

  /**
   * Asserts workflow help text appears.
   *
   * @param array $data
   *   The node data for the test.
   * @param bool $is_workflow_admin
   *   Indicates the user can administer workflows.
   */
  private function assertWorkflowText(array $data, bool $is_workflow_admin = FALSE) {
    $node = $data['node'];
    if ($is_workflow_admin && $data['invalid']) {
      $this->assertSession()->pageTextContains('Approvals may only be assigned when new revisions will not change the current state of the content.');
      $this->assertSession()->pageTextContains(sprintf('The current %s moderation state does not support this action and may need a new content transition.',
        $node->get('moderation_state')->getString()));
      $this->assertSession()->linkExists('a new content transition');
      $link = $this->workflow->toUrl('edit-form')->toString();
      $this->assertSession()->linkByHrefExists($link);
    }
    else {
      $this->assertSession()->pageTextNotContains('Approvals may only be assigned when new revisions will not change the current state of the content.');
      $this->assertSession()->pageTextNotContains(sprintf('The current %s moderation state does not support this action and may need a new content transition.',
        $node->get('moderation_state')->getString()));
      $this->assertSession()->linkNotExists('a new content transition');
      $link = $this->workflow->toUrl('edit-form')->toString();
      $this->assertSession()->linkByHrefNotExists($link);
    }
  }

  /**
   * Asserts the default published content message.
   *
   * @param array $data
   *   The node data for the test.
   */
  private function assertPublishedContent(array $data) {
    if ($data['published']) {
      $this->assertSession()->pageTextContains('The current version is published and cannot be approved further.');
    }
    else {
      $this->assertSession()->pageTextNotContains('The current version is published and cannot be approved further.');
    }
  }

  /**
   * Asserts the review requested message.
   *
   * @param array $data
   *   The node data for the test.
   * @param bool $is_admin
   *   Indicates the user is an administrator.
   * @param bool $expected
   *   Indicates the user was assigned as a reviewer.
   */
  private function assertReviewRequested(array $data, bool $is_admin = FALSE, bool $expected = TRUE) {
    $node = $data['node'];
    $no_review = [
      $this->plainNode->id(),
      $this->publishedNode->id(),
      $this->archivedNode->id(),
    ];
    if (in_array($node->id(), $no_review, TRUE)) {
      if ($is_admin) {
        $this->assertSession()->pageTextContains('You may assign approvals for this content.');
      }
      else {
        $this->assertSession()->pageTextContains('You may view the approvals for this content.');
        $this->assertSession()->pageTextNotContains('You may assign approvals for this content.');
      }
    }
    else {
      if ($data['active'] && $expected) {
        $this->assertSession()->pageTextContains('Your review has been requested on the current revision.');
        $this->assertSession()->linkExists('current revision');
      }
      else {
        $this->assertSession()->pageTextNotContains('Your review has been requested on the current revision.');
        $this->assertSession()->linkNotExists('current revision');
      }
    }
  }

  /**
   * Asserts the add approval button for admins.
   *
   * @param array $data
   *   The node data for the test.
   * @param bool $is_admin
   *   Indicates the user can assign approvals.
   */
  private function assertAddApproval(array $data, bool $is_admin = FALSE) {
    if (!$is_admin || $data['published'] || $data['invalid']) {
      $this->assertSession()->buttonNotExists('Add approval');
    }
    else {
      $this->assertSession()->buttonExists('Add approval');
    }
  }

  /**
   * Asserts the content of operations links.
   *
   * @param array $data
   *   The node data for the test.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The user viewing the page.
   * @param bool $is_assigned
   *   Indicates the user has an assigned approval.
   * @param bool $is_admin
   *   Indicates the user can assign approvals.
   */
  private function assertOperationsLinks(array $data, AccountInterface $account, bool $is_assigned = FALSE, bool $is_admin = FALSE) {
    $node = $data['node'];

    $approvers = $this->approverManager->getApprovalRequests($node);

    foreach ($approvers as $index => $approver) {
      // Each of these roles can see notes.
      $link = '/approver/' . $approver->id() . '/notes';
      if ($approver->get('notes')->value !== NULL) {
        $this->assertSession()->linkExists('View notes', $index);
        $this->assertSession()->linkByHrefExists($link);
      }
      else {
        $this->assertSession()->linkByHrefNotExists($link);
      }

      // Admins and approvers can see revisions.
      // This link URL may appear elsewhere on the page.
      if ($is_assigned || $is_admin) {
        $this->assertSession()->linkExists('Show revision', $index);
      }

      // Admins and approvers can approve and request changes.
      $link = '/approver/' . $approver->id() . '/approve';
      $link2 = '/approver/' . $approver->id() . '/reject';
      if (($is_assigned && $approver->getApproverId() == $account->id()) || $is_admin) {
        // Assigned approvers only see one link.
        $item = $is_admin ? $index : 0;
        $this->assertSession()->linkExists('Approve revision', $item);
        $this->assertSession()->linkByHrefExists($link);
        $this->assertSession()->linkExists('Request changes', $item);
        $this->assertSession()->linkByHrefExists($link2);
      }
      else {
        $this->assertSession()->linkByHrefNotExists($link);
        $this->assertSession()->linkByHrefNotExists($link2);
      }

      // Admins can edit and delete.
      $link = '/approver/' . $approver->id() . '/edit';
      $link2 = '/approver/' . $approver->id() . '/delete';
      if ($is_admin) {
        $this->assertSession()->linkExists('Edit request', $index);
        $this->assertSession()->linkByHrefExists($link);
        $this->assertSession()->linkExists('Delete request', $index);
        $this->assertSession()->linkByHrefExists($link2);
      }
      else {
        $this->assertSession()->linkNotExists('Edit request');
        $this->assertSession()->linkByHrefNotExists($link);
        $this->assertSession()->linkNotExists('Delete request');
        $this->assertSession()->linkByHrefNotExists($link2);
      }
    }
  }

  /**
   * Asserts the content of active approvals view.
   *
   * @param array $data
   *   The node data for the test.
   */
  private function assertActiveView(array $data) {
    if ($data['active']) {
      $node = $data['node'];
      $approvers = $this->approverManager->getApprovalRequests($node);

      $approver_labels = $this->xpath('//td[contains(@class, "views-field-approver-uid")]');
      $required_values = $this->xpath('//td[contains(@class, "views-field-required")]');
      $creation_dates = $this->xpath('//td[contains(@class, "views-field-created")]');
      $deadlines = $this->xpath('//td[contains(@class, "views-field-deadline")]');
      $statuses = $this->xpath('//td[contains(@class, "views-field-status")]');
      $assigners = $this->xpath('//td[contains(@class, "views-field-uid")]');

      $delta = 0;
      foreach ($approvers as $approver) {
        $this->assertEquals(trim($approver_labels[$delta]->getText()), $approver->getApproverLabel());

        $required = (bool) $approver->get('required')->value ? 'Yes' : 'No';
        $this->assertEquals(trim($required_values[$delta]->getText()), $required);

        $created = date('d-M-Y', $approver->get('created')->value);
        $this->assertEquals(trim($creation_dates[$delta]->getText()), $created);

        $deadline = date('d-M-Y', $approver->get('deadline')->value);
        $this->assertEquals(trim($deadlines[$delta]->getText()), $deadline);

        $status = $approver->getStatusName();
        $this->assertEquals(trim($statuses[$delta]->getText()), $status);

        $this->assertEquals(trim($assigners[$delta]->getText()), $approver->getAssignerLabel());

        $delta++;
      }
    }
  }

  /**
   * Asserts the content of prior approvals view.
   *
   * @param array $data
   *   The node data for the test.
   */
  private function assertPriorView(array $data) {
    if ($data['prior']) {
      $node = $data['node'];
      $parameters = [
        'content_entity_type_id' => 'node',
        'content_entity_id' => $node->id(),
        'workflow_id' => $this->workflow->id(),
      ];
      $unfiltered_approvers = $this->approverStorage->loadByProperties($parameters);
      $approvers = [];
      foreach ($unfiltered_approvers as $approver) {
        assert($approver instanceof ApproverInterface);
        if ($approver->get('content_entity_revision_id')->value != $node->getRevisionId()) {
          $approvers[] = $approver;
        }
      }

      // @todo We can currently only test items *not* in the active view.
      $revision_ids = $this->xpath('//td[contains(@class, "views-field-content-entity-revision-id")]');
      $moderation_states = $this->xpath('//td[contains(@class, "views-field-moderation-state")]');

      $delta = 0;
      foreach ($approvers as $approver) {
        $this->assertEquals(trim($revision_ids[$delta]->getText()), $approver->get('content_entity_revision_id')->value);
        $this->assertEquals(trim($moderation_states[$delta]->getText()), $approver->get('moderation_state')->getString());
        $delta++;
      }
    }
  }

}
