<?php

namespace Drupal\Tests\workbench_approver\Functional;

use Drupal\workbench_approver\Entity\ApproverInterface;

/**
 * Tests that the Content Moderation form appears as expected.
 *
 * @group workbench_approver
 */
class ContentModerationIntegrationTest extends WorkbenchApproverTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'content_moderation',
    'workbench_approver',
  ];

  /**
   * A node with assignments.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $assignedNode;

  /**
   * A node without assignments.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $nonAssignedNode;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $values = [
      'type' => 'approved',
      'title' => 'Assigned node',
      'moderation_state' => 'published',
    ];
    $this->assignedNode = $this->drupalCreateNode($values);
    $this->assignedNode = $this->createRevision($this->assignedNode);

    // Assign two users to approve this node.
    // Use a fixed timestamp for testing (Wed Feb 02 2035 02:00:37 GMT+0000)
    $approver_values = [
      'deadline' => 2053994437,
    ];
    $this->addApprovalRequest($this->adminUser, $this->assignedNode, $approver_values);
    $this->addApprovalRequest($this->editorUser, $this->assignedNode, $approver_values);

    $values = [
      'type' => 'approved',
      'title' => 'Non-assigned node',
      'moderation_state' => 'published',
    ];
    $this->nonAssignedNode = $this->drupalCreateNode($values);
    $this->nonAssignedNode = $this->createRevision($this->nonAssignedNode);
  }

  /**
   * Tests the embedded preview form.
   *
   * @covers \Drupal\workbench_approver\Services\ApproverStateTransitionValidation
   */
  public function testContentModerationForm() {
    $path = '/node/' . $this->assignedNode->id() . '/latest';
    $path2 = '/node/' . $this->nonAssignedNode->id() . '/latest';

    // Check our storage.
    $this->checkSetupValues();

    // The latest version page should not show the moderation form
    // because no revision states are available.
    $this->drupalLogin($this->adminUser);
    $this->drupalGet($path);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextNotContains('Moderation state');
    $this->assertSession()->pageTextNotContains('Change to');
    $this->assertSession()->pageTextNotContains('Log message');
    $this->assertSession()->pageTextNotContains('Draft');
    $this->assertSession()->fieldNotExists('edit-new-state');
    $this->assertSession()->buttonNotExists('Apply');

    // The latest version page should show the moderation form and have "Draft"
    // status, because the pending revision is in "Draft".
    $this->drupalGet($path2);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Moderation state');
    $this->assertSession()->pageTextContains('Change to');
    $this->assertSession()->pageTextContains('Log message');
    $this->assertSession()->pageTextContains('Draft');
    $this->assertSession()->fieldExists('edit-new-state');
    $this->assertSession()->buttonExists('Apply');
  }

  /**
   * Tests the node edit form.
   *
   * @covers \Drupal\workbench_approver\Services\ApproverStateTransitionValidation
   */
  public function testEditForm() {
    $path = '/node/' . $this->assignedNode->id() . '/edit';
    $path2 = '/node/' . $this->nonAssignedNode->id() . '/edit';

    // Check our storage.
    $this->checkSetupValues();

    // This node should be draft and not allow change.
    $this->drupalLogin($this->adminUser);
    $this->drupalGet($path);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Current state Draft');
    $this->assertSession()->optionExists('moderation_state[0][state]', 'draft');
    $this->assertSession()->optionNotExists('moderation_state[0][state]', 'published');
    $this->assertSession()->optionNotExists('moderation_state[0][state]', 'archived');

    // This node should be draft and should allow change.
    $this->drupalGet($path2);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Current state Draft');
    $this->assertSession()->optionExists('moderation_state[0][state]', 'draft');
    $this->assertSession()->optionExists('moderation_state[0][state]', 'published');

    // Explicit inner service test; see ContentModerationValidationInnerTest.
    $this->createRevision($this->nonAssignedNode, 'published');
    $this->drupalGet($path2);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Current state Published');
    $this->assertSession()->optionExists('moderation_state[0][state]', 'draft');
    $this->assertSession()->optionExists('moderation_state[0][state]', 'published');
    $this->assertSession()->optionExists('moderation_state[0][state]', 'archived');
  }

  /**
   * Checks that our test conditions are set as expected.
   */
  public function checkSetupValues() {
    $this->assertEquals('draft', $this->assignedNode->get('moderation_state')->getString());
    $this->assertEquals('draft', $this->nonAssignedNode->get('moderation_state')->getString());

    $approvers = $this->approverStorage->loadByProperties([
      'content_entity_id' => $this->assignedNode->id(),
      'content_entity_revision_id' => $this->assignedNode->getRevisionId(),
      'status' => ApproverInterface::APPROVAL_PENDING,
    ]);
    $this->assertEquals(count($approvers), 2);

    $approvers = $this->approverStorage->loadByProperties([
      'content_entity_id' => $this->nonAssignedNode->id(),
      'content_entity_revision_id' => $this->nonAssignedNode->getRevisionId(),
    ]);
    $this->assertEquals(count($approvers), 0);
  }

}
