<?php

namespace Drupal\Tests\workbench_approver\Functional;

/**
 * Tests for access to approval pages and tabs.
 *
 * @group workbench_approver
 */
class IsApprovableTest extends WorkbenchApproverTestBase {

  /**
   * The state transition validator.
   *
   * @var \Drupal\workbench_approver\Services\ApproverStateTransitionValidation
   */
  protected $transitionValidation;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    // Add the validator service.
    $this->transitionValidation = $this->container->get('content_moderation.state_transition_validation');
  }

  /**
   * Tests that approvals require valid transitions.
   *
   * @covers \Drupal\workbench_approver\Services\ApproverStateTransitionValidation::isApprovable
   * @covers \Drupal\workbench_approver\Controller\ApproverController::approvalTab
   */
  public function testIsApprovable() {
    $values = [
      'type' => 'approved',
      'title' => 'Approved node',
      'moderation_state' => 'published',
    ];
    $approved_node = $this->drupalCreateNode($values);
    $path = '/node/' . $approved_node->id() . '/approver';

    $this->drupalLogin($this->adminUser);

    // The default workflow allows draft-to-draft and published-to-published.
    // However, we explicitly disallow approvals of published content.
    $is_approvable = $this->transitionValidation->isApprovable($approved_node);
    $this->assertFalse($is_approvable, 'Content does not need approval');

    // Check content on the approve tab.
    $this->drupalGet($path);
    $this->assertSession()->pageTextContains('The current version is published and cannot be approved further.');

    // Create a draft.
    $approved_revision = $this->createRevision($approved_node, 'draft');
    $is_approvable = $this->transitionValidation->isApprovable($approved_revision);
    $this->assertTrue($is_approvable, 'Content requires approval');

    // Check content on the approve tab.
    $this->drupalGet($path);
    $this->assertSession()->pageTextNotContains('The current version is published and cannot be approved further.');
    $this->assertSession()->pageTextNotContains('Approvals may only be assigned when new revisions will not change the current state of the content.');
    $this->assertSession()->pageTextNotContains('No current approvals are required for this content.');
    $this->assertSession()->buttonExists('Add approval');

    // Publish the revision.
    $published_revision = $this->createRevision($approved_revision, 'published');
    // Check content on the approve tab.
    $this->drupalGet($path);
    $this->assertSession()->pageTextContains('The current version is published and cannot be approved further.');

    // Create an archive.
    $archived_revision = $this->createRevision($published_revision, 'archived');
    $is_approvable = $this->transitionValidation->isApprovable($archived_revision);
    // Check content on the approve tab.
    $this->assertFalse($is_approvable, 'Content not eligible for approval');
    $this->drupalGet($path);
    $this->assertSession()->pageTextContains('No current approvals are required for this content.');

    // Check message text for non-admins.
    $this->drupalLogin($this->editorUser);
    $this->drupalGet($path);
    $this->assertSession()->pageTextContains('No current approvals are required for this content.');
    $this->assertSession()->buttonNotExists('Add approval');

    // Check message text for workflow admins.
    $this->drupalLogin($this->workflowAdminUser);
    $this->drupalGet($path);
    $this->assertSession()->pageTextContains('Approvals may only be assigned when new revisions will not change the current state of the content.');
  }

}
