<?php

namespace Drupal\Tests\workbench_approver\Functional;

use Drupal\node\NodeInterface;
use Drupal\workbench_approver\Entity\ApproverInterface;

/**
 * Tests approval handling when revisions are updated.
 *
 * @group workbench_approver
 */
class RevisionRevertTest extends WorkbenchApproverTestBase {

  /**
   * Tests when an older draft is made the current draft.
   *
   * All approvers should be marked outdated.
   */
  public function testRevisionRevert() {
    $values = [
      'type' => 'approved',
      'title' => 'Approved node',
      'moderation_state' => 'draft',
    ];
    $draft_node = $this->drupalCreateNode($values);
    $this->assertEquals(1, $draft_node->getRevisionId());

    // Create a published revision.
    $published_node = $this->createRevision($draft_node, 'published');
    assert($published_node instanceof NodeInterface);
    $this->assertEquals(2, $published_node->getRevisionId());

    // Create a new revision and assign approvers.
    $revised_node = $this->createRevision($published_node);
    assert($revised_node instanceof NodeInterface);
    $this->assertEquals(3, $revised_node->getRevisionId());
    $this->addApprovalRequest($this->adminUser, $revised_node);
    $this->addApprovalRequest($this->editorUser, $revised_node);

    // Assert creation of approvals on the new draft.
    $this->assertApprovers($revised_node, 2);

    // Check that the original draft does not have approvals.
    $this->assertApprovers($draft_node, 0);

    // Show approvals for debugging.
    $this->loadTestPaths($draft_node);

    // Revert to $draft_node.
    $this->revertToDraft($draft_node);

    // Show approvals for debugging.
    $this->loadTestPaths($draft_node);

    // Get the revision created by the revert action.
    $reverted_node = $this->getLatestNodeRevision($draft_node);

    // @todo Clear cache -- this should happen but does not.
    $this->approverStorage->resetCache();

    // The new revision cannot have approval requests.
    $this->assertApprovers($reverted_node, 0);

    // Check that approvers were reassigned.
    $approvers = $this->assertApprovers($revised_node, 2);
    foreach ($approvers as $approver) {
      $this->assertEquals(ApproverInterface::APPROVAL_OUTDATED, $approver->getStatus());
    }

    // The old revision still has no requests.
    $this->assertApprovers($draft_node, 0);

    // Show approvals for debugging.
    $this->loadTestPaths($draft_node);
  }

  /**
   * Tests reverting to a draft after one user requested changes.
   *
   * No users should be reassigned to the new draft.
   *
   * Users who approved the draft should be marked invalidated.
   */
  public function testRevisionChangesRequested() {
    $values = [
      'type' => 'approved',
      'title' => 'Approved node',
      'moderation_state' => 'draft',
    ];
    $draft_node = $this->drupalCreateNode($values);
    $this->assertEquals(1, $draft_node->getRevisionId());

    // Create a published revision.
    $published_node = $this->createRevision($draft_node, 'published');
    assert($published_node instanceof NodeInterface);
    $this->assertEquals(2, $published_node->getRevisionId());

    // Create a new revision and assign approvers.
    $revised_node = $this->createRevision($published_node);
    assert($revised_node instanceof NodeInterface);
    $this->assertEquals(3, $revised_node->getRevisionId());
    $this->addApprovalRequest($this->adminUser, $revised_node, ['status' => ApproverInterface::APPROVAL_GRANTED]);
    $this->addApprovalRequest($this->editorUser, $revised_node, ['status' => ApproverInterface::APPROVAL_REJECTED]);

    // Assert creation of approvals on the new draft.
    $approvers = $this->assertApprovers($revised_node, 2);
    foreach ($approvers as $approver) {
      if ($approver->getApproverId() == $this->editorUser->id()) {
        $this->assertEquals(ApproverInterface::APPROVAL_REJECTED, $approver->getStatus());
      }
      else {
        $this->assertEquals(ApproverInterface::APPROVAL_GRANTED, $approver->getStatus());
      }
    }

    // Check that the original draft does not have approvals.
    $this->assertApprovers($draft_node, 0);

    // Show approvals for debugging.
    $this->loadTestPaths($draft_node);

    // Revert to $draft_node.
    $this->revertToDraft($draft_node);

    // Show approvals for debugging.
    $this->loadTestPaths($draft_node);

    // Get the revision created by the revert action.
    $reverted_node = $this->getLatestNodeRevision($draft_node);

    // @todo Clear cache -- this should happen but does not.
    $this->approverStorage->resetCache();

    // The new revision cannot have approval requests.
    $this->assertApprovers($reverted_node, 0);

    // Check that approvals were reassigned.
    $approvers = $this->assertApprovers($revised_node, 2);
    foreach ($approvers as $approver) {
      if ($approver->getApproverId() == $this->editorUser->id()) {
        $this->assertEquals(ApproverInterface::APPROVAL_REJECTED, $approver->getStatus());
      }
      else {
        $this->assertEquals(ApproverInterface::APPROVAL_GRANTED, $approver->getStatus());
      }
    }

    // The old revision still has no requests.
    $this->assertApprovers($draft_node, 0);

    // Show approvals for debugging.
    $this->loadTestPaths($draft_node);
  }

  /**
   * Reverts a node to a previous revision.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node under test.
   *
   * @return void
   *   No return value.
   */
  private function revertToDraft(NodeInterface $node): void {
    $this->drupalLogin($this->adminUser);
    $revert_path = '/node/' . $node->id() . '/revisions/' . $node->getRevisionId() . '/revert';
    $this->drupalGet($revert_path);
    $this->assertSession()->buttonExists('Revert')->click();
  }

}
