<?php

namespace Drupal\Tests\workbench_approver\Functional;

use Drupal\node\NodeInterface;
use Drupal\workbench_approver\Entity\ApproverInterface;

/**
 * Tests for cases when content revisions are made.
 *
 * @group workbench_approver
 */
class TransitionTest extends WorkbenchApproverTestBase {

  /**
   * Tests making a new draft from a draft.
   *
   * All users should be reassigned to the new draft.
   */
  public function testNewDraft() {
    $node = $this->createNodeWithApprovalRequests();

    // Save and get a revision.
    $node_revision = $this->createRevision($node);

    // Show approvals and revisions for debugging.
    $this->loadTestPaths($node);

    // Test the revised requests.
    $approvers = $this->assertApprovers($node_revision, 2);
    foreach ($approvers as $approver) {
      $this->assertEquals(ApproverInterface::APPROVAL_PENDING, $approver->getStatus());
    }

    // Now test the original requests.
    $approvers = $this->assertApprovers($node, 2);
    foreach ($approvers as $approver) {
      $this->assertEquals(ApproverInterface::APPROVAL_OUTDATED, $approver->getStatus());
    }

    // Show approvals and revisions for debugging.
    $this->loadTestPaths($node);
  }

  /**
   * Tests making a new draft after one user requested changes.
   *
   * All users should be reassigned to the new draft.
   *
   * Users who approved the draft should maintain the approved status.
   */
  public function testChangesRequested() {
    $node = $this->createNodeWithApprovalRequests();
    $this->setApprovals($node);

    // Save and get a revision.
    $node_revision = $this->createRevision($node);

    // Show approvals and revisions for debugging.
    $this->loadTestPaths($node);

    // Test the revised requests. There should be only 1 created request now.
    $approvers = $this->assertApprovers($node_revision, 1);
    foreach ($approvers as $approver) {
      $this->assertEquals(ApproverInterface::APPROVAL_PENDING, $approver->getStatus());
    }

    // Now test the original requests.
    $approvers = $this->assertApprovers($node, 2);
    foreach ($approvers as $approver) {
      if ($approver->getApproverId() == $this->editorUser->id()) {
        $this->assertEquals(ApproverInterface::APPROVAL_REJECTED, $approver->getStatus());
      }
      else {
        $this->assertEquals(ApproverInterface::APPROVAL_GRANTED, $approver->getStatus());
      }
    }
  }

  /**
   * Tests transitioning to published after all users approve.
   *
   * Users who approved should show approved status.
   */
  public function testRevisionApproved() {
    $node = $this->createNodeWithApprovalRequests();
    $approvers = $this->approverManager->getApprovalRequests($node);
    foreach ($approvers as $approver) {
      $approver->set('status', ApproverInterface::APPROVAL_GRANTED)->save();
    }

    // Show approvals and revisions for debugging.
    $this->drupalLogin($this->adminUser);
    $this->loadTestPaths($node);

    // Save and get a revision.
    $node_revision = $this->createRevision($node, 'published');

    // Show approvals and revisions for debugging.
    $this->loadTestPaths($node);

    // Test the revised requests.
    $this->assertApprovers($node_revision, 0);

    // Now test the original requests.
    $approvers = $this->assertApprovers($node, 2);
    foreach ($approvers as $approver) {
      $this->assertEquals(ApproverInterface::APPROVAL_GRANTED, $approver->getStatus());
    }
  }

  /**
   * Tests transitioning all required users approve.
   *
   * Users who approved should show approved status.
   *
   * Users who requested changes should show invalidated status.
   */
  public function testChangesNotRequired() {
    $node = $this->createNodeWithApprovalRequests(NULL, []);
    $approvers = $this->approverManager->getApprovalRequests($node);
    foreach ($approvers as $approver) {
      if ($approver->getApproverId() == $this->editorUser->id()) {
        $approver->set('required', FALSE);
        $approver->set('status', ApproverInterface::APPROVAL_REJECTED)->save();
      }
      else {
        $approver->set('status', ApproverInterface::APPROVAL_GRANTED)->save();
      }
    }

    // Show approvals and revisions for debugging.
    $this->drupalLogin($this->adminUser);
    $this->loadTestPaths($node);

    // Save and get a revision.
    $node_revision = $this->createRevision($node, 'published');

    // Show approvals and revisions for debugging.
    $this->loadTestPaths($node);

    // Test the revised requests.
    $this->assertApprovers($node_revision, 0);

    // Now test the original requests.
    $approvers = $this->assertApprovers($node, 2);
    foreach ($approvers as $approver) {
      if ($approver->getApproverId() == $this->editorUser->id()) {
        $this->assertEquals(ApproverInterface::APPROVAL_REJECTED, $approver->getStatus());
      }
      else {
        $this->assertEquals(ApproverInterface::APPROVAL_GRANTED, $approver->getStatus());
      }
    }
  }

  /**
   * Given a node with two approvers, set one yes and one no.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node revision.
   */
  public function setApprovals(NodeInterface $node) {
    $approvers = $this->approverManager->getApprovalRequests($node);

    foreach ($approvers as $approver) {
      if ($approver->getApproverId() == $this->editorUser->id()) {
        $approver->set('status', ApproverInterface::APPROVAL_REJECTED)->save();
      }
      else {
        $approver->set('status', ApproverInterface::APPROVAL_GRANTED)->save();
      }
    }

    $approvers = $this->assertApprovers($node, 2);
    foreach ($approvers as $approver) {
      if ($approver->getApproverId() == $this->editorUser->id()) {
        $this->assertEquals(ApproverInterface::APPROVAL_REJECTED, $approver->getStatus());
      }
      else {
        $this->assertEquals(ApproverInterface::APPROVAL_GRANTED, $approver->getStatus());
      }
    }
  }

}
