<?php

namespace Drupal\Tests\workbench_approver\Functional;

use Drupal\user\Entity\User;
use Drupal\workbench_approver\Entity\ApproverInterface;

/**
 * Tests for content on the admin by content list.
 *
 * @group workbench_approver
 */
class UserDeletedTest extends WorkbenchApproverTestBase {

  /**
   * Tests user hooks (for deleted users).
   *
   * @covers ::workbench_approver_user_predelete()
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function testApprovalsWithDeletedApproverMethod1() {
    $this->approvalsWithDeletedApprover('user_cancel_delete');
  }

  /**
   * Tests user hooks (for deleted users).
   *
   * @covers ::workbench_approver_user_predelete()
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function testApprovalsWithDeletedApproverMethod2() {
    $this->approvalsWithDeletedApprover('user_cancel_reassign');
  }

  /**
   * Tests user hooks (for deleted users).
   *
   * @covers ::workbench_approver_user_predelete()
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function testApprovalsWithDeletedApproverAndSameOwnerMethod1() {
    $this->approvalsWithDeletedApproverAndSameOwner('user_cancel_delete');
  }

  /**
   * Tests user hooks (for deleted users).
   *
   * @covers ::workbench_approver_user_predelete()
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function testApprovalsWithDeletedApproverAndSameOwnerMethod2() {
    $this->approvalsWithDeletedApproverAndSameOwner('user_cancel_reassign');
  }

  /**
   * Helps to test deleted approvers for approvals.
   *
   * Needed this helper method because with the two methods for deletion in a
   * foreach loop the user ids where mixed in the notes messages and weird
   * errors appeared.
   *
   * @param string $delete_cancel_method
   *   The delete method to be tested.
   *
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function approvalsWithDeletedApprover(string $delete_cancel_method): void {
    // Select blocking method.
    $this->config('user.settings')->set('cancel_method', $delete_cancel_method)->save();

    // Create users to delete in different tests.
    $approverToDelete = $this->createUserWithAssignedRole('editor');

    $approved_node_1 = $this->drupalCreateNode([
      'type' => 'approved',
      'title' => 'Approved node 1',
      'moderation_state' => 'draft',
      'uid' => $this->editorUser,
    ]);

    $approved_node_2 = $this->drupalCreateNode([
      'type' => 'approved',
      'title' => 'Approved node 2',
      'moderation_state' => 'draft',
      'uid' => $this->editorUser,
    ]);

    // Create revisions and approvals.
    $this->drupalLogin($this->adminUser);

    /** @var \Drupal\Core\Entity\ContentEntityInterface $node_revision_1 */
    $node_revision_1 = $this->createRevision($approved_node_1);
    /** @var \Drupal\Core\Entity\ContentEntityInterface $node_revision_2 */
    $node_revision_2 = $this->createRevision($approved_node_2);

    $this->addApprovalRequest($approverToDelete, $node_revision_1, [
      'status' => ApproverInterface::APPROVAL_PENDING,
      'uid' => $this->editorUser,
    ]);
    $this->addApprovalRequest($approverToDelete, $node_revision_2, [
      'status' => ApproverInterface::APPROVAL_PENDING,
      'uid' => $this->editorUser,
    ]);

    // Delete approver.
    $this->drupalGet('user/' . $approverToDelete->id() . '/cancel');
    $this->assertSession()->pageTextContains('Cancellation method');

    // Confirm user deletion.
    $this->submitForm([], 'Confirm');
    // Confirm account cancellation request.
    $this->assertNull(User::load($approverToDelete->id()), 'User found in the database but not expected.');

    // Confirm changes in approvals.
    $approval_ids = $this->approverManager->getApprovalRequestIdsByEntityKeys('owner', $this->editorUser->id());
    $this->assertNotEmpty($approval_ids);
    $approvals = $this->approverStorage->loadMultiple($approval_ids);
    $this->assertNotEmpty($approvals);
    /** @var \Drupal\workbench_approver\Entity\Approver $approval */
    foreach ($approvals as $approval) {
      $this->assertSame(ApproverInterface::APPROVAL_RESCINDED, $approval->getStatus(), 'Status of the Approval not changed as expected for blocked user.');
      $this->assertSame("The status of this request has been changed to \"{$approval->getStatusName()}\". The approver of this request with username \"{$approverToDelete->getAccountName()}\" has been deleted.",
        strip_tags($approval->get('notes')->getString()), 'The notes message in the approval is not the expected one.');
      $this->assertEquals($this->editorUser->id(), $approval->getApproverId(), 'The approver was not changed to the owner as expected.');
    }
  }

  /**
   * Helps to test deleted approvers for approvals.
   *
   * Needed this helper method because with the two methods for deletion in a
   * foreach loop the user ids where mixed in the notes messages and weird
   * errors appeared.
   *
   * @param string $delete_cancel_method
   *   The delete method to be tested.
   *
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function approvalsWithDeletedApproverAndSameOwner(string $delete_cancel_method): void {
    // Select blocking method.
    $this->config('user.settings')->set('cancel_method', $delete_cancel_method)->save();

    // Create users to delete in different tests.
    $approverAndOwnerToDelete = $this->createUserWithAssignedRole('editor');

    $approved_node_1 = $this->drupalCreateNode([
      'type' => 'approved',
      'title' => 'Approved node 1',
      'moderation_state' => 'draft',
      'uid' => $this->editorUser,
    ]);

    // Create revisions and approvals.
    $this->drupalLogin($this->adminUser);

    /** @var \Drupal\Core\Entity\ContentEntityInterface $node_revision_1 */
    $node_revision_1 = $this->createRevision($approved_node_1);

    $this->addApprovalRequest($approverAndOwnerToDelete, $node_revision_1, [
      'status' => ApproverInterface::APPROVAL_PENDING,
      'uid' => $approverAndOwnerToDelete,
    ]);

    // Delete approver.
    $this->drupalGet('user/' . $approverAndOwnerToDelete->id() . '/cancel');
    $this->assertSession()->pageTextContains('Cancellation method');

    // Confirm user deletion.
    $this->submitForm([], 'Confirm');
    // Confirm account cancellation request.
    $this->assertNull(User::load($approverAndOwnerToDelete->id()), 'User found in the database but not expected.');

    // Confirm the approval was also deleted.
    $approval_ids = $this->approverManager->getApprovalRequestIdsByEntityKeys('owner', $approverAndOwnerToDelete->id());
    $this->assertEmpty($approval_ids);
  }

  /**
   * Tests user hooks (for deleted users).
   *
   * @covers ::workbench_approver_user_predelete()
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function testApprovalsWithDeletedOwnerMethod1() {
    $this->approvalsWithDeletedOwner('user_cancel_delete');
  }

  /**
   * Tests user hooks (for deleted users).
   *
   * @covers ::workbench_approver_user_predelete()
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function testApprovalsWithDeletedOwnerMethod2() {
    $this->approvalsWithDeletedOwner('user_cancel_reassign');
  }

  /**
   * Helps to test deleted owners for approvals.
   *
   * Needed this helper method because with the two methods for deletion in a
   * foreach loop the user ids where mixed in the notes messages and weird
   * errors appeared.
   *
   * @param string $delete_cancel_method
   *   The delete method to be tested.
   *
   * @throws \Behat\Mink\Exception\ResponseTextException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function approvalsWithDeletedOwner(string $delete_cancel_method): void {
    // Select blocking method.
    $this->config('user.settings')->set('cancel_method', $delete_cancel_method)->save();

    // Create user to delete in different tests.
    $ownerToDelete = $this->createUserWithAssignedRole('editor');

    $approved_node_1 = $this->drupalCreateNode([
      'type' => 'approved',
      'title' => 'Approved node 1',
      'moderation_state' => 'draft',
      'uid' => $this->editorUser,
    ]);

    $approved_node_2 = $this->drupalCreateNode([
      'type' => 'approved',
      'title' => 'Approved node 2',
      'moderation_state' => 'draft',
      'uid' => $this->editorUser,
    ]);

    // Create revisions and approvals.
    $this->drupalLogin($this->adminUser);

    /** @var \Drupal\Core\Entity\ContentEntityInterface $node_revision_1 */
    $node_revision_1 = $this->createRevision($approved_node_1);
    /** @var \Drupal\Core\Entity\ContentEntityInterface $node_revision_2 */
    $node_revision_2 = $this->createRevision($approved_node_2);

    $this->addApprovalRequest($this->editorUser, $node_revision_1, [
      'status' => ApproverInterface::APPROVAL_PENDING,
      'uid' => $ownerToDelete,
    ]);
    $this->addApprovalRequest($this->editorUser, $node_revision_2, [
      'status' => ApproverInterface::APPROVAL_PENDING,
      'uid' => $ownerToDelete,
    ]);

    // Delete approver.
    $this->drupalGet('user/' . $ownerToDelete->id() . '/cancel');
    $this->assertSession()->pageTextContains('Cancellation method');

    // Confirm user deletion.
    $this->submitForm([], 'Confirm');
    // Confirm account cancellation request.
    $this->assertNull(User::load($ownerToDelete->id()), 'User found in the database but not expected.');

    // Confirm changes in approvals.
    $approval_ids = $this->approverManager->getApprovalRequestIdsByEntityKeys('uid', $ownerToDelete->id());
    $this->assertEmpty($approval_ids);
    $approval_ids = $this->approverManager->getApprovalRequestIdsByEntityKeys('approver_uid', $this->editorUser->id());
    $this->assertNotEmpty($approval_ids);
    $approvals = $this->approverStorage->loadMultiple($approval_ids);
    $this->assertNotEmpty($approvals);
    /** @var \Drupal\workbench_approver\Entity\Approver $approval */
    foreach ($approvals as $approval) {
      $this->assertEquals(0, $approval->getOwnerId(), 'The owner was not changed as expected.');
      $this->assertSame("The owner of this request with username \"{$ownerToDelete->getAccountName()}\" has been deleted. The new owner of this approval is now the \"Anonymous\" user (0).",
        strip_tags($approval->get('notes')->getString()), 'The notes message in the approval is not the expected one.');
    }
  }

}
