<?php

namespace Drupal\Tests\workbench_approver\Functional;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\node\NodeInterface;
use Drupal\node\NodeStorageInterface;
use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\content_moderation\Traits\ContentModerationTestTrait;
use Drupal\Tests\workbench_approver\Traits\WorkbenchApproverTestTrait;
use Drupal\user\UserInterface;
use Drupal\workbench_approver\ApproverStorageInterface;
use Drupal\workbench_approver\Entity\ApproverInterface;
use Drupal\workbench_approver\Services\ApproverManagerInterface;
use Drupal\workflows\WorkflowInterface;

/**
 * Common setup tasks for Workbench Approver tests.
 */
abstract class WorkbenchApproverTestBase extends BrowserTestBase {

  use ContentModerationTestTrait;
  use WorkbenchApproverTestTrait;

  /**
   * Disable config schema checking.
   *
   * The views dependencies for the module throw errors.
   *
   * @var bool
   */
  protected $strictConfigSchema = FALSE; // phpcs:ignore

  /**
   * The default theme.
   *
   * @var string
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'workbench_approver',
  ];

  /**
   * Admin user.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $adminUser;

  /**
   * Editor user.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $editorUser;

  /**
   * An additional user with the 'editor' role.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $anotherEditorUser;

  /**
   * Observer user.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $observerUser;

  /**
   * Workflow admin user.
   *
   * @var \Drupal\user\UserInterface
   */
  protected UserInterface $workflowAdminUser;

  /**
   * The editorial workflow entity.
   *
   * @var \Drupal\workflows\WorkflowInterface
   */
  protected WorkflowInterface $workflow;

  /**
   * The Approver manager service.
   *
   * @var \Drupal\workbench_approver\Services\ApproverManagerInterface|null
   */
  public ?ApproverManagerInterface $approverManager;

  /**
   * The Approver storage service.
   *
   * @var \Drupal\workbench_approver\ApproverStorageInterface
   */
  public ApproverStorageInterface $approverStorage;

  /**
   * The Node storage service.
   *
   * @var \Drupal\node\NodeStorageInterface
   */
  public NodeStorageInterface $nodeStorage;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->defaultApproverTestSetup();
  }

  /**
   * Handles setting up our test values in dependency order.
   */
  public function defaultApproverTestSetup() {
    // Our test content types (`approved` and `non_approved`).
    $this->setUpContentTypes();

    // Setup elements from the ContentModerationTestTrait.
    $this->workflow = $this->createEditorialWorkflow();
    $this->addEntityTypeAndBundleToWorkflow($this->workflow, 'node', 'approved');

    // Add our users.
    $this->setUpDefaultUsers();

    // Set up the manager services.
    $this->approverManager = $this->container->get('workbench_approver.manager');
    $this->nodeStorage = $this->container->get('entity_type.manager')->getStorage('node');
    $this->approverStorage = $this->container->get('entity_type.manager')->getStorage('approver');
  }

  /**
   * Sets up our default test accounts.
   *
   * Note that this must come after node type and workflow
   * creation, due to permission dependencies.
   */
  public function setUpDefaultUsers(): void {
    $this->adminUser = $this->setUpAdministratorUser();
    $this->editorUser = $this->setUpEditorialUser();
    $this->anotherEditorUser = $this->createUserWithAssignedRole('editor');
    $this->observerUser = $this->setUpObserverUser();
    $this->workflowAdminUser = $this->setUpWorkflowAdminUser();
  }

  /**
   * Makes common assertions used in multiple tests.
   *
   * Here we test that our arguments to the approver manager return
   * only the users that we expect.
   *
   * @param array $arguments
   *   The arguments passed to getPotentialApproverIds().
   * @param array $expected
   *   An array of uids that should be returned.
   * @param array $not_expected
   *   An array of uids that should not be returned.
   */
  public function checkApprovers(array $arguments, array $expected, array $not_expected): void {
    $potential = $this->approverManager->getPotentialApproverIds($arguments);
    // Use array_values() since keys may not match and do not matter.
    $this->assertEquals(array_values($expected), array_values($potential), 'Expected users match');
    $this->assertEmpty(array_intersect($not_expected, $potential));
  }

  /**
   * Creates a node and a revision and then assigns approvers.
   *
   * @param string|null $title
   *   An optional title.
   * @param array $users
   *   An optional array of users.
   * @param array $options
   *   Any options to pass to the approval request.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface
   *   The last created revision.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function createNodeWithApprovalRequests(string $title = NULL, array $users = [], array $options = []): ContentEntityInterface {
    $values = [
      'type' => 'approved',
      'title' => $title ?? 'Approved node',
      'moderation_state' => 'published',
    ];
    $node = $this->drupalCreateNode($values);

    if (count($users) === 0) {
      $users = [$this->adminUser, $this->editorUser];
    }
    // Create a revision and assign approvers.
    /** @var \Drupal\Core\Entity\ContentEntityInterface $node_revision */
    $node_revision = $this->createRevision($node);
    foreach ($users as $user) {
      $this->addApprovalRequest($user, $node_revision, $options);
    }

    // Assert creation.
    $approvers = $this->approverManager->getApprovalRequests($node_revision);
    $this->assertEquals(count($users), count($approvers));

    return $node_revision;
  }

  /**
   * Given a node with two approvers, set one yes and one no.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node revision.
   */
  public function setApprovals(NodeInterface $node) {
    $approvers = $this->approverManager->getApprovalRequests($node);

    foreach ($approvers as $approver) {
      if ($approver->getApproverId() == $this->editorUser->id()) {
        $approver->set('status', ApproverInterface::APPROVAL_REJECTED)->save();
      }
      else {
        $approver->set('status', ApproverInterface::APPROVAL_GRANTED)->save();
      }
    }

    $approvers = $this->approverManager->getApprovalRequests($node);
    $this->assertEquals(2, count($approvers));
    foreach ($approvers as $approver) {
      if ($approver->getApproverId() == $this->editorUser->id()) {
        $this->assertEquals(ApproverInterface::APPROVAL_REJECTED, $approver->getStatus());
      }
      else {
        $this->assertEquals(ApproverInterface::APPROVAL_GRANTED, $approver->getStatus());
      }
    }
  }

  /**
   * Gets the latest revision of a node.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node under test.
   *
   * @return \Drupal\Core\Entity\RevisionableInterface|null
   *   The latest node revision.
   */
  public function getLatestNodeRevision(NodeInterface $node) {
    // @todo Clear cache -- this should happen but does not.
    $this->nodeStorage->resetCache([$node->id()]);

    $latest_nid = $this->nodeStorage->getLatestRevisionId($node->id());
    $latest_node = $this->nodeStorage->loadRevision($latest_nid);
    // Ensure that we have a different revision than the $node.
    $this->assertNotEquals($latest_node->getRevisionId(), $node->getRevisionId());
    $this->assertEquals(4, $latest_node->getRevisionId());

    return $latest_node;
  }

  /**
   * Add permissions to a given role.
   *
   * Useful if you need extra permissions for a role but only during a
   * specific test.
   *
   * @param array $permissions
   *   An array of permissions to add to the role.
   * @param string $role
   *   The role to be modified.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function addPermissionsToRole(array $permissions, string $role): void {
    $role_storage = $this->container->get('entity_type.manager')->getStorage('user_role');
    /** @var \Drupal\user\Entity\Role $role */
    $role = $role_storage->load($role);
    foreach ($permissions as $permission) {
      $role->grantPermission($permission);
    }
    $role->save();
  }

}
