<?php

namespace Drupal\Tests\workbench_approver\Functional;

use Drupal\Core\Entity\EntityInterface;

/**
 * Tests for the Workbench tab.
 *
 * @group workbench_approver
 */
class WorkbenchIntegrationTest extends WorkbenchApproverTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'workbench',
    'workbench_approver',
  ];

  /**
   * Test access to the tab.
   */
  public function testAccessToWorkbench() {
    $path = '/admin/workbench/content/approvals';

    // Currently no users have access to Workbench.
    $users = [
      NULL,
      $this->adminUser,
      $this->editorUser,
      $this->observerUser,
      $this->workflowAdminUser,
    ];

    foreach ($users as $user) {
      if ($user instanceof EntityInterface) {
        $this->drupalLogin($user);
      }
      $this->drupalGet($path);
      $this->assertSession()->statusCodeEquals(403);
    }

    $this->drupalLogout();
    $this->addPermissionForTest();

    // Test again.
    foreach ($users as $user) {
      $code = 403;
      if ($user instanceof EntityInterface) {
        $this->drupalLogin($user);
        $code = $user->hasPermission('access workbench') ? 200 : 403;
      }
      $this->drupalGet($path);
      $this->assertSession()->statusCodeEquals($code);
    }
  }

  /**
   * Tests content on the tab.
   */
  public function testWorkbenchContent() {
    $path = '/admin/workbench/content/approvals';
    $this->addPermissionForTest();

    $nodes = [];
    for ($i = 1; $i <= 5; $i++) {
      $nodes[$i] = $this->createNodeWithApprovalRequests("Test node $i");
    }

    $this->drupalLogin($this->editorUser);

    $this->drupalGet($path);

    $node_labels = $this->xpath('//td[contains(@class, "views-field-workbench-approver-latest-revision")]');
    $required_values = $this->xpath('//td[contains(@class, "views-field-required")]');
    $statuses = $this->xpath('//td[contains(@class, "views-field-status")]');
    $creation_dates = $this->xpath('//td[contains(@class, "views-field-created")]');
    $deadlines = $this->xpath('//td[contains(@class, "views-field-deadline")]');
    $assigners = $this->xpath('//td[contains(@class, "views-field-uid")]');

    $delta = 0;
    foreach ($nodes as $node) {
      $approver = $this->approverManager->getApprovalRequestByUser($node, $this->editorUser);

      $this->assertEquals(trim($node_labels[$delta]->getText()), $node->label());

      $required = (bool) $approver->get('required')->value ? 'Yes' : 'No';
      $this->assertEquals(trim($required_values[$delta]->getText()), $required);

      $created = date('d-M-Y', $approver->get('created')->value);
      $this->assertEquals(trim($creation_dates[$delta]->getText()), $created);

      $deadline = date('d-M-Y', $approver->get('deadline')->value);
      $this->assertEquals(trim($deadlines[$delta]->getText()), $deadline);

      $status = $approver->getStatusName();
      $this->assertEquals(trim($statuses[$delta]->getText()), $status);

      $this->assertEquals(trim($assigners[$delta]->getText()), $approver->getAssignerLabel());

      $delta++;
    }
    $headers = [
      'Title',
      'Required',
      'Status',
      'Due date',
      'Requested',
      'Requested by',
      'Operations',
    ];
    foreach ($headers as $header) {
      $this->assertSession()->pageTextContains($header);
    }

    // Now make one node published. It should not appear.
    $this->createRevision($nodes[5], 'published');

    // Delete one node. It should not appear.
    $nodes[4]->delete();

    // Delete one approval request. It should not appear.
    $approver = $this->approverManager->getApprovalRequestByUser($nodes[3], $this->editorUser);
    $approver->delete();

    $this->drupalGet($path);

    foreach ($nodes as $id => $node) {
      if ($id < 3) {
        $this->assertSession()->pageTextContains($node->label());
      }
      else {
        $this->assertSession()->pageTextNotContains($node->label());
      }
    }

  }

  /**
   * Adds permission to use workbench.
   */
  private function addPermissionForTest() {
    // Add permissions for editor and observer roles.
    $rids = [
      'editor',
      'observer',
    ];
    $roles = $this->container->get('entity_type.manager')->getStorage('user_role')->loadMultiple($rids);
    foreach ($roles as $role) {
      $this->grantPermissions($role, ['access workbench']);
    }
  }

}
