<?php

namespace Drupal\Tests\workbench_approver\Traits;

use Drupal\Core\Entity\EntityInterface;
use Drupal\node\NodeInterface;
use Drupal\user\UserInterface;
use Drupal\workbench_approver\Entity\ApproverInterface;

/**
 * Contains helper classes for tests to set up various configuration.
 *
 * These methods are in a Trait rather than in WorkbenchApproverTestBase
 * so that they can be called from Kernel tests.
 *
 * Names of some methods are verbose to prevent collision with
 * \Drupal\Tests\workbench_access\Traits\WorkbenchAccessTestTrait.
 */
trait WorkbenchApproverTestTrait {

  /**
   * Sets up two content types for testing.
   *
   * 'approved' content is subject to the 'editorial' workflow.
   * 'non_approved' content is not subject to the 'editorial' workflow.
   */
  public function setUpContentTypes():void {
    $this->createContentType(['type' => 'approved']);
    $this->createContentType(['type' => 'non_approved']);
  }

  /**
   * Sets up a user with an editor role that has access to content.
   *
   * @return \Drupal\user\UserInterface
   *   The user entity.
   */
  public function setUpEditorialUser() {
    $editor_rid = $this->createRole([
      'access administration pages',
      'create approved content',
      'create non_approved content',
      'edit any approved content',
      'edit any non_approved content',
      'use editorial transition create_new_draft',
      'use workbench approver',
      'view workbench approver information',
    ], 'editor');

    return $this->createUserWithAssignedRole($editor_rid);
  }

  /**
   * Sets up a user with an observer role that has access to content.
   *
   * @return \Drupal\user\UserInterface
   *   The user entity.
   */
  public function setUpObserverUser() {
    $observer_rid = $this->createRole([
      'access administration pages',
      'view workbench approver information',
    ], 'observer');

    return $this->createUserWithAssignedRole($observer_rid);
  }

  /**
   * Sets up a user with an admin role.
   *
   * @param array $additional_permissions
   *   Array of additional permissions beyond 'access administration pages' and
   *   'assign workbench access'.
   *
   * @return \Drupal\user\UserInterface
   *   The user entity.
   */
  protected function setUpAdministratorUser(array $additional_permissions = []) {
    $admin_rid = $this->createRole(array_merge($additional_permissions, [
      'access administration pages',
      'administer users',
      'assign workbench approver',
      'administer workbench approver',
      'create approved content',
      'edit any approved content',
      'delete any approved content',
      'create non_approved content',
      'edit any non_approved content',
      'delete all revisions',
      'revert all revisions',
      'view all revisions',
      'use editorial transition create_new_draft',
      'use editorial transition publish',
      'use editorial transition archive',
      'use editorial transition archived_draft',
      'use editorial transition archived_published',
      'use workbench approver',
      'view workbench approver information',
      'view latest version',
      'view any unpublished content',
    ]), 'admin');

    return $this->createUserWithAssignedRole($admin_rid);
  }

  /**
   * Sets up a user with an workflow admin role.
   *
   * @param array $additional_permissions
   *   Array of additional permissions beyond 'access administration pages' and
   *   'assign workbench access'.
   *
   * @return \Drupal\user\UserInterface
   *   The user entity.
   */
  protected function setUpWorkflowAdminUser(array $additional_permissions = []) {
    $admin_rid = $this->createRole(array_merge($additional_permissions, [
      'access administration pages',
      'assign workbench approver',
      'administer workbench approver',
      'administer workflows',
      'create approved content',
      'edit any approved content',
      'create non_approved content',
      'edit any non_approved content',
      'use editorial transition create_new_draft',
      'use editorial transition publish',
      'use editorial transition archive',
      'use editorial transition archived_draft',
      'use editorial transition archived_published',
      'use workbench approver',
      'view workbench approver information',
      'view latest version',
      'view any unpublished content',
    ]), 'workflow_admin');

    return $this->createUserWithAssignedRole($admin_rid);
  }

  /**
   * Sets up a user with a given role and saves it.
   *
   * @param string $rid
   *   The role id.
   *
   * @return \Drupal\user\UserInterface
   *   The user entity.
   */
  protected function createUserWithAssignedRole($rid) {
    $user = $this->createUser();
    $user->addRole($rid);
    $user->save();

    return $user;
  }

  /**
   * Creates a revision of an existing entity.
   *
   * If you intend to use the revision as a test
   * variable, make sure the caller sets a variable
   * to the return of this method.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param string $to_state
   *   The target moderation state.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   The revised entity.
   */
  public function createRevision(EntityInterface $entity, $to_state = 'draft'): EntityInterface {
    // If we don't clone here, it updates the caller by reference.
    $revision = clone $entity;
    assert($revision instanceof NodeInterface);
    $revision->set('moderation_state', $to_state);
    $revision->setNewRevision(TRUE);
    $revision->save();

    // Something is hanging the route parameter during testing.
    \Drupal::service('router.builder')->rebuildIfNeeded();

    return $revision;
  }

  /**
   * Adds an approver to an entity.
   *
   * @param \Drupal\user\UserInterface $account
   *   The user to add. This is a UserInterface because of how
   *   createUserWithAssignedRole() behaves.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The target entity.
   * @param array $values
   *   Any non-default values to set on the approver entity.
   *
   * @return \Drupal\workbench_approver\Entity\ApproverInterface
   *   The created approval entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException|\Drupal\Core\Entity\EntityStorageException
   */
  public function addApprovalRequest(UserInterface $account, EntityInterface $entity, array $values = []): ApproverInterface {
    $storage = \Drupal::entityTypeManager()->getStorage('approver');

    $time = \Drupal::time()->getRequestTime();

    assert($entity instanceof NodeInterface);
    $default_values = [
      'workflow_id' => $this->workflow->id(),
      'moderation_state' => $entity->get('moderation_state')->getString(),
      'content_entity_type_id' => $entity->getEntityTypeId(),
      'content_entity_id' => $entity->id(),
      'content_entity_revision_id' => $entity->getRevisionId(),
      'uid' => $this->adminUser->id(),
      'status' => ApproverInterface::APPROVAL_PENDING,
      'required' => TRUE,
      'approver_uid' => $account->id(),
      'deadline' => $time + (24 * 7 * 3600),
      'notes' => '',
      'created' => $time,
      'changed' => $time,
      'langcode' => $entity->language()->getId(),
    ];

    $new_values = array_replace($default_values, $values);

    /** @var \Drupal\workbench_approver\Entity\ApproverInterface $approver */
    $approver = $storage->create($new_values);
    $approver->save();

    return $approver;
  }

  /**
   * Tests common paths for approval handling.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node.
   *
   * @return void
   *   No return value.
   */
  public function loadTestPaths(NodeInterface $node): void {
    $approvals_path = '/node/' . $node->id() . '/approver';
    $revisions_path = '/node/' . $node->id() . '/revisions';
    // Show approvals for debugging.
    $this->drupalGet($approvals_path);
    $this->drupalGet($revisions_path);
  }

  /**
   * Assert our approval request count.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node revision under test.
   * @param int $count
   *   The expected number of approval requests.
   *
   * @return \Drupal\workbench_approver\Entity\ApproverInterface[]
   *   An array of approval requests.
   */
  public function assertApprovers(NodeInterface $node, int $count) {
    $approvers = $this->approverManager->getApprovalRequests($node);
    $this->assertEquals($count, count($approvers));

    return $approvers;
  }

  /**
   * Tests that form paths are accessible as expected.
   *
   * @param array $paths
   *   An array where the keys are paths and the value is an
   *   array of allowed user ids.
   * @param array $users
   *   An array of users under test.
   *
   * @return void
   *   No return value.
   */
  public function assertApprovalPathAccess(array $paths, array $users): void {
    if ($this->loggedInUser) {
      $this->drupalLogout();
    }

    foreach ($users as $user) {
      if ($user !== NULL) {
        $this->drupalLogin($user);
      }
      foreach ($paths as $path => $expected) {
        $this->drupalGet($path);
        if ($user !== NULL && in_array($user->id(), $expected, TRUE)) {
          $this->assertSession()->statusCodeEquals(200);
        }
        else {
          $this->assertSession()->statusCodeEquals(403);
        }
      }
    }
  }

}
